"""
Manage groups on FreeBSD

.. important::
    If you feel that Salt should be using this module to manage groups on a
    minion, and it is using a different module (or gives an error similar to
    *'group.info' is not available*), see :ref:`here
    <module-provider-override>`.
"""

import logging

import salt.utils.args
import salt.utils.data

try:
    import grp
except ImportError:
    pass

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "group"


def __virtual__():
    """
    Set the user module if the kernel is FreeBSD or Dragonfly
    """
    if __grains__["kernel"] in ("FreeBSD", "DragonFly"):
        return __virtualname__
    return (
        False,
        "The pw_group execution module cannot be loaded: system is not supported.",
    )


def add(name, gid=None, **kwargs):
    """
    .. versionchanged:: 3006.0

    Add the specified group

    name
        Name of the new group

    gid
        Use GID for the new group

    CLI Example:

    .. code-block:: bash

        salt '*' group.add foo 3456
    """
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    if salt.utils.data.is_true(kwargs.pop("system", False)):
        log.warning("pw_group module does not support the 'system' argument")
    if "non_unique" in kwargs:
        log.warning("The non_unique parameter is not supported on this platform.")
    if kwargs:
        log.warning("Invalid kwargs passed to group.add")

    cmd = "pw groupadd "
    if gid:
        cmd += f"-g {gid} "
    cmd = f"{cmd} -n {name}"
    ret = __salt__["cmd.run_all"](cmd, python_shell=False)

    return not ret["retcode"]


def delete(name):
    """
    Remove the named group

    CLI Example:

    .. code-block:: bash

        salt '*' group.delete foo
    """
    ret = __salt__["cmd.run_all"](f"pw groupdel {name}", python_shell=False)

    return not ret["retcode"]


def info(name):
    """
    Return information about a group

    CLI Example:

    .. code-block:: bash

        salt '*' group.info foo
    """
    try:
        grinfo = grp.getgrnam(name)
    except KeyError:
        return {}
    else:
        return {
            "name": grinfo.gr_name,
            "passwd": grinfo.gr_passwd,
            "gid": grinfo.gr_gid,
            "members": grinfo.gr_mem,
        }


def getent(refresh=False):
    """
    Return info on all groups

    CLI Example:

    .. code-block:: bash

        salt '*' group.getent
    """
    if "group.getent" in __context__ and not refresh:
        return __context__["group.getent"]

    ret = []
    for grinfo in grp.getgrall():
        ret.append(info(grinfo.gr_name))
    __context__["group.getent"] = ret
    return ret


def chgid(name, gid):
    """
    Change the gid for a named group

    CLI Example:

    .. code-block:: bash

        salt '*' group.chgid foo 4376
    """
    pre_gid = __salt__["file.group_to_gid"](name)
    if gid == pre_gid:
        return True
    cmd = f"pw groupmod {name} -g {gid}"
    __salt__["cmd.run"](cmd, python_shell=False)
    post_gid = __salt__["file.group_to_gid"](name)
    if post_gid != pre_gid:
        return post_gid == gid
    return False


def adduser(name, username):
    """
    Add a user in the group.

    CLI Example:

    .. code-block:: bash

         salt '*' group.adduser foo bar

    Verifies if a valid username 'bar' as a member of an existing group 'foo',
    if not then adds it.
    """
    # Note: pw exits with code 65 if group is unknown
    retcode = __salt__["cmd.retcode"](
        f"pw groupmod {name} -m {username}", python_shell=False
    )

    return not retcode


def deluser(name, username):
    """
    Remove a user from the group.

    CLI Example:

    .. code-block:: bash

         salt '*' group.deluser foo bar

    Removes a member user 'bar' from a group 'foo'. If group is not present
    then returns True.
    """
    grp_info = __salt__["group.info"](name)

    if username not in grp_info["members"]:
        return True

    # Note: pw exits with code 65 if group is unknown
    retcode = __salt__["cmd.retcode"](
        f"pw groupmod {name} -d {username}", python_shell=False
    )

    return not retcode


def members(name, members_list):
    """
    Replaces members of the group with a provided list.

    .. versionadded:: 2015.5.4

    CLI Example:

    .. code-block:: bash

        salt '*' group.members foo 'user1,user2,user3,...'

    Replaces a membership list for a local group 'foo'.
        foo:x:1234:user1,user2,user3,...
    """

    retcode = __salt__["cmd.retcode"](
        f"pw groupmod {name} -M {members_list}", python_shell=False
    )

    return not retcode



import codecs
import glob
import logging
import os
import textwrap

import pytest

import salt.loader
import salt.template
import salt.utils.data
import salt.utils.files
import salt.utils.reactor as reactor
import salt.utils.yaml
from tests.support.mock import MagicMock, Mock, mock_open, patch

log = logging.getLogger(__name__)

REACTOR_CONFIG = """\
reactor:
  - old_runner:
    - /srv/reactor/old_runner.sls
  - old_wheel:
    - /srv/reactor/old_wheel.sls
  - old_local:
    - /srv/reactor/old_local.sls
  - old_cmd:
    - /srv/reactor/old_cmd.sls
  - old_caller:
    - /srv/reactor/old_caller.sls
  - new_runner:
    - /srv/reactor/new_runner.sls
  - new_wheel:
    - /srv/reactor/new_wheel.sls
  - new_local:
    - /srv/reactor/new_local.sls
  - new_cmd:
    - /srv/reactor/new_cmd.sls
  - new_caller:
    - /srv/reactor/new_caller.sls
"""

REACTOR_DATA = {
    "runner": {"data": {"message": "This is an error"}},
    "wheel": {"data": {"id": "foo"}},
    "local": {"data": {"pkg": "zsh", "repo": "updates"}},
    "cmd": {"data": {"pkg": "zsh", "repo": "updates"}},
    "caller": {"data": {"path": "/tmp/foo"}},
}

SLS = {
    "/srv/reactor/old_runner.sls": textwrap.dedent(
        """\
        raise_error:
          runner.error.error:
            - name: Exception
            - message: {{ data['data']['message'] }}
        """
    ),
    "/srv/reactor/old_wheel.sls": textwrap.dedent(
        """\
        remove_key:
          wheel.key.delete:
            - match: {{ data['data']['id'] }}
        """
    ),
    "/srv/reactor/old_local.sls": textwrap.dedent(
        """\
        install_zsh:
          local.state.single:
            - tgt: test
            - arg:
              - pkg.installed
              - {{ data['data']['pkg'] }}
            - kwarg:
                fromrepo: {{ data['data']['repo'] }}
        """
    ),
    "/srv/reactor/old_cmd.sls": textwrap.dedent(
        """\
        install_zsh:
          cmd.state.single:
            - tgt: test
            - arg:
              - pkg.installed
              - {{ data['data']['pkg'] }}
            - kwarg:
                fromrepo: {{ data['data']['repo'] }}
        """
    ),
    "/srv/reactor/old_caller.sls": textwrap.dedent(
        """\
        touch_file:
          caller.file.touch:
            - args:
              - {{ data['data']['path'] }}
        """
    ),
    "/srv/reactor/new_runner.sls": textwrap.dedent(
        """\
        raise_error:
          runner.error.error:
            - args:
              - name: Exception
              - message: {{ data['data']['message'] }}
        """
    ),
    "/srv/reactor/new_wheel.sls": textwrap.dedent(
        """\
        remove_key:
          wheel.key.delete:
            - args:
              - match: {{ data['data']['id'] }}
        """
    ),
    "/srv/reactor/new_local.sls": textwrap.dedent(
        """\
        install_zsh:
          local.state.single:
            - tgt: test
            - args:
              - fun: pkg.installed
              - name: {{ data['data']['pkg'] }}
              - fromrepo: {{ data['data']['repo'] }}
        """
    ),
    "/srv/reactor/new_cmd.sls": textwrap.dedent(
        """\
        install_zsh:
          cmd.state.single:
            - tgt: test
            - args:
              - fun: pkg.installed
              - name: {{ data['data']['pkg'] }}
              - fromrepo: {{ data['data']['repo'] }}
        """
    ),
    "/srv/reactor/new_caller.sls": textwrap.dedent(
        """\
        touch_file:
          caller.file.touch:
            - args:
              - name: {{ data['data']['path'] }}
        """
    ),
}

LOW_CHUNKS = {
    "old_runner": [
        {
            "state": "runner",
            "__id__": "raise_error",
            "__sls__": "/srv/reactor/old_runner.sls",
            "order": 1,
            "fun": "error.error",
            "name": "Exception",
            "message": "This is an error",
        }
    ],
    "old_wheel": [
        {
            "state": "wheel",
            "__id__": "remove_key",
            "name": "remove_key",
            "__sls__": "/srv/reactor/old_wheel.sls",
            "order": 1,
            "fun": "key.delete",
            "match": "foo",
        }
    ],
    "old_local": [
        {
            "state": "local",
            "__id__": "install_zsh",
            "name": "install_zsh",
            "__sls__": "/srv/reactor/old_local.sls",
            "order": 1,
            "tgt": "test",
            "fun": "state.single",
            "arg": ["pkg.installed", "zsh"],
            "kwarg": {"fromrepo": "updates"},
        }
    ],
    "old_cmd": [
        {
            "state": "local",
            "__id__": "install_zsh",
            "name": "install_zsh",
            "__sls__": "/srv/reactor/old_cmd.sls",
            "order": 1,
            "tgt": "test",
            "fun": "state.single",
            "arg": ["pkg.installed", "zsh"],
            "kwarg": {"fromrepo": "updates"},
        }
    ],
    "old_caller": [
        {
            "state": "caller",
            "__id__": "touch_file",
            "name": "touch_file",
            "__sls__": "/srv/reactor/old_caller.sls",
            "order": 1,
            "fun": "file.touch",
            "args": ["/tmp/foo"],
        }
    ],
    "new_runner": [
        {
            "state": "runner",
            "__id__": "raise_error",
            "name": "raise_error",
            "__sls__": "/srv/reactor/new_runner.sls",
            "order": 1,
            "fun": "error.error",
            "args": [{"name": "Exception"}, {"message": "This is an error"}],
        }
    ],
    "new_wheel": [
        {
            "state": "wheel",
            "__id__": "remove_key",
            "name": "remove_key",
            "__sls__": "/srv/reactor/new_wheel.sls",
            "order": 1,
            "fun": "key.delete",
            "args": [{"match": "foo"}],
        }
    ],
    "new_local": [
        {
            "state": "local",
            "__id__": "install_zsh",
            "name": "install_zsh",
            "__sls__": "/srv/reactor/new_local.sls",
            "order": 1,
            "tgt": "test",
            "fun": "state.single",
            "args": [
                {"fun": "pkg.installed"},
                {"name": "zsh"},
                {"fromrepo": "updates"},
            ],
        }
    ],
    "new_cmd": [
        {
            "state": "local",
            "__id__": "install_zsh",
            "name": "install_zsh",
            "__sls__": "/srv/reactor/new_cmd.sls",
            "order": 1,
            "tgt": "test",
            "fun": "state.single",
            "args": [
                {"fun": "pkg.installed"},
                {"name": "zsh"},
                {"fromrepo": "updates"},
            ],
        }
    ],
    "new_caller": [
        {
            "state": "caller",
            "__id__": "touch_file",
            "name": "touch_file",
            "__sls__": "/srv/reactor/new_caller.sls",
            "order": 1,
            "fun": "file.touch",
            "args": [{"name": "/tmp/foo"}],
        }
    ],
}

WRAPPER_CALLS = {
    "old_runner": (
        "error.error",
        {
            "__state__": "runner",
            "__id__": "raise_error",
            "__sls__": "/srv/reactor/old_runner.sls",
            "__user__": "Reactor",
            "order": 1,
            "arg": [],
            "kwarg": {"name": "Exception", "message": "This is an error"},
            "name": "Exception",
            "message": "This is an error",
        },
    ),
    "old_wheel": (
        "key.delete",
        {
            "__state__": "wheel",
            "__id__": "remove_key",
            "name": "remove_key",
            "__sls__": "/srv/reactor/old_wheel.sls",
            "order": 1,
            "__user__": "Reactor",
            "arg": ["foo"],
            "kwarg": {},
            "match": "foo",
        },
    ),
    "old_local": {
        "args": ("test", "state.single"),
        "kwargs": {
            "state": "local",
            "__id__": "install_zsh",
            "name": "install_zsh",
            "__sls__": "/srv/reactor/old_local.sls",
            "order": 1,
            "arg": ["pkg.installed", "zsh"],
            "kwarg": {"fromrepo": "updates"},
        },
    },
    "old_cmd": {
        "args": ("test", "state.single"),
        "kwargs": {
            "state": "local",
            "__id__": "install_zsh",
            "name": "install_zsh",
            "__sls__": "/srv/reactor/old_cmd.sls",
            "order": 1,
            "arg": ["pkg.installed", "zsh"],
            "kwarg": {"fromrepo": "updates"},
        },
    },
    "old_caller": {"args": ("file.touch", "/tmp/foo"), "kwargs": {}},
    "new_runner": (
        "error.error",
        {
            "__state__": "runner",
            "__id__": "raise_error",
            "name": "raise_error",
            "__sls__": "/srv/reactor/new_runner.sls",
            "__user__": "Reactor",
            "order": 1,
            "arg": (),
            "kwarg": {"name": "Exception", "message": "This is an error"},
        },
    ),
    "new_wheel": (
        "key.delete",
        {
            "__state__": "wheel",
            "__id__": "remove_key",
            "name": "remove_key",
            "__sls__": "/srv/reactor/new_wheel.sls",
            "order": 1,
            "__user__": "Reactor",
            "arg": (),
            "kwarg": {"match": "foo"},
        },
    ),
    "new_local": {
        "args": ("test", "state.single"),
        "kwargs": {
            "state": "local",
            "__id__": "install_zsh",
            "name": "install_zsh",
            "__sls__": "/srv/reactor/new_local.sls",
            "order": 1,
            "arg": (),
            "kwarg": {"fun": "pkg.installed", "name": "zsh", "fromrepo": "updates"},
        },
    },
    "new_cmd": {
        "args": ("test", "state.single"),
        "kwargs": {
            "state": "local",
            "__id__": "install_zsh",
            "name": "install_zsh",
            "__sls__": "/srv/reactor/new_cmd.sls",
            "order": 1,
            "arg": (),
            "kwarg": {"fun": "pkg.installed", "name": "zsh", "fromrepo": "updates"},
        },
    },
    "new_caller": {"args": ("file.touch",), "kwargs": {"name": "/tmp/foo"}},
}


# -----------------------------------------------------------------------------
# FIXTURES
# -----------------------------------------------------------------------------
@pytest.fixture
def react_master_opts(master_opts):
    opts = {
        # Minimal stand-in for a real master config
        "file_roots": {"base": []},
        "renderer": "jinja|yaml",
    }
    master_opts.update(opts)
    # Optionally parse the reactor config for convenience
    reactor_config = salt.utils.yaml.safe_load(REACTOR_CONFIG)
    master_opts.update(reactor_config)
    return master_opts


@pytest.fixture
def test_reactor(react_master_opts):
    """
    Create a Reactor instance for testing
    """
    return reactor.Reactor(react_master_opts)


@pytest.fixture
def reaction_map(react_master_opts):
    """
    Reaction map from the configured reactor
    """
    return salt.utils.data.repack_dictlist(react_master_opts["reactor"])


@pytest.fixture
def render_pipe(react_master_opts):
    """
    Render pipeline
    """
    renderers = salt.loader.render(react_master_opts, {})
    return [(renderers[x], "") for x in ("jinja", "yaml")]


# -----------------------------------------------------------------------------
# TESTS for Reactor building the low chunks
# -----------------------------------------------------------------------------
@pytest.mark.parametrize("schema", ["old", "new"])
@pytest.mark.parametrize("rtype", list(REACTOR_DATA.keys()))
def test_reactor_reactions(schema, rtype, test_reactor, render_pipe):
    """
    Ensure correct reactions are built from the configured SLS files and tag data.
    """
    tag = f"{schema}_{rtype}"
    reactors_list = test_reactor.list_reactors(tag)

    # Patch out globbing since these SLS files don't actually exist on disk
    with patch.object(glob, "glob", MagicMock(side_effect=lambda x: [x])):
        with patch.object(os.path, "isfile", MagicMock(return_value=True)):
            with patch.object(
                salt.utils.files, "is_empty", MagicMock(return_value=False)
            ):
                with patch.object(
                    codecs, "open", mock_open(read_data=SLS[reactors_list[0]])
                ):
                    with patch.object(
                        salt.template,
                        "template_shebang",
                        MagicMock(return_value=render_pipe),
                    ):
                        reactions = test_reactor.reactions(
                            tag, REACTOR_DATA[rtype], reactors_list
                        )
    assert reactions == LOW_CHUNKS[tag], f"Reactions did not match for tag: {tag}"


def test_list_reactors(test_reactor, reaction_map):
    """
    Ensure list_reactors() returns the correct list of reactor SLS files for each tag.
    """
    for schema in ("old", "new"):
        for rtype in REACTOR_DATA:
            tag = f"{schema}_{rtype}"
            assert test_reactor.list_reactors(tag) == reaction_map[tag]


# -----------------------------------------------------------------------------
# FIXTURE for Reactor Wrap
# -----------------------------------------------------------------------------
@pytest.fixture
def react_wrap(react_master_opts):
    """
    Create a ReactWrap instance
    """
    return reactor.ReactWrap(react_master_opts)


# -----------------------------------------------------------------------------
# TESTS for ReactWrap
# -----------------------------------------------------------------------------
@pytest.mark.parametrize("schema", ["old", "new"])
def test_runner(schema, react_wrap):
    """
    Test runner reactions using both the old and new config schema
    """
    tag = f"{schema}_runner"
    chunk = LOW_CHUNKS[tag][0]
    thread_pool = Mock()
    thread_pool.fire_async = Mock()
    with patch.object(react_wrap, "pool", thread_pool):
        react_wrap.run(chunk)
    thread_pool.fire_async.assert_called_with(
        react_wrap.client_cache["runner"].low,
        args=WRAPPER_CALLS[tag],
    )


@pytest.mark.parametrize("schema", ["old", "new"])
def test_wheel(schema, react_wrap):
    """
    Test wheel reactions using both the old and new config schema
    """
    tag = f"{schema}_wheel"
    chunk = LOW_CHUNKS[tag][0]
    thread_pool = Mock()
    thread_pool.fire_async = Mock()
    with patch.object(react_wrap, "pool", thread_pool):
        react_wrap.run(chunk)
    thread_pool.fire_async.assert_called_with(
        react_wrap.client_cache["wheel"].low,
        args=WRAPPER_CALLS[tag],
    )


@pytest.mark.parametrize("schema", ["old", "new"])
def test_local(schema, react_wrap):
    """
    Test local reactions using both the old and new config schema
    """
    tag = f"{schema}_local"
    chunk = LOW_CHUNKS[tag][0]
    client_cache = {"local": Mock()}
    client_cache["local"].cmd_async = Mock()
    with patch.object(react_wrap, "client_cache", client_cache):
        react_wrap.run(chunk)
    client_cache["local"].cmd_async.assert_called_with(
        *WRAPPER_CALLS[tag]["args"], **WRAPPER_CALLS[tag]["kwargs"]
    )


@pytest.mark.parametrize("schema", ["old", "new"])
def test_cmd(schema, react_wrap):
    """
    Test cmd reactions (alias for 'local') using both the old and new config schema
    """
    tag = f"{schema}_cmd"
    chunk = LOW_CHUNKS[tag][0]
    client_cache = {"local": Mock()}
    client_cache["local"].cmd_async = Mock()
    with patch.object(react_wrap, "client_cache", client_cache):
        react_wrap.run(chunk)
    client_cache["local"].cmd_async.assert_called_with(
        *WRAPPER_CALLS[tag]["args"], **WRAPPER_CALLS[tag]["kwargs"]
    )


@pytest.mark.parametrize("schema", ["old", "new"])
def test_caller(schema, react_wrap):
    """
    Test caller reactions using both the old and new config schema
    """
    tag = f"{schema}_caller"
    chunk = LOW_CHUNKS[tag][0]
    client_cache = {"caller": Mock()}
    client_cache["caller"].cmd = Mock()
    with patch.object(react_wrap, "client_cache", client_cache):
        react_wrap.run(chunk)
    client_cache["caller"].cmd.assert_called_with(
        *WRAPPER_CALLS[tag]["args"], **WRAPPER_CALLS[tag]["kwargs"]
    )


@pytest.mark.parametrize("file_client", ["runner", "wheel"])
def test_client_cache_missing_key(file_client, react_wrap):
    """
    Test client_cache file_client missing, gets repopulated
    """
    client_cache = {}
    tag = f"new_{file_client}"
    chunk = LOW_CHUNKS[tag][0]
    with patch.object(react_wrap, "client_cache", client_cache):
        if f"{file_client}" == "runner":
            react_wrap.runner(chunk)
        elif f"{file_client}" == "wheel":
            react_wrap.wheel(chunk)
        else:
            # catch need for new check
            assert f"{file_client}" == "bad parameterization"

        file_client_key = None
        for key in react_wrap.client_cache.keys():
            if key == f"{file_client}":
                file_client_key = key

        assert file_client_key == f"{file_client}"



"""
An execution module which can manipulate an f5 bigip via iControl REST
    :maturity:      develop
    :platform:      f5_bigip_11.6
"""

import salt.exceptions
import salt.utils.json

try:
    import requests
    import requests.exceptions

    HAS_LIBS = True
except ImportError:
    HAS_LIBS = False


# Define the module's virtual name
__virtualname__ = "bigip"


def __virtual__():
    """
    Only return if requests is installed
    """
    if HAS_LIBS:
        return __virtualname__
    return (
        False,
        "The bigip execution module cannot be loaded: "
        "python requests library not available.",
    )


BIG_IP_URL_BASE = "https://{host}/mgmt/tm"


def _build_session(username, password, trans_label=None):
    """
    Create a session to be used when connecting to iControl REST.
    """

    bigip = requests.session()
    bigip.auth = (username, password)
    bigip.verify = True
    bigip.headers.update({"Content-Type": "application/json"})

    if trans_label:
        # pull the trans id from the grain
        trans_id = __salt__["grains.get"](f"bigip_f5_trans:{trans_label}")

        if trans_id:
            bigip.headers.update({"X-F5-REST-Coordination-Id": trans_id})
        else:
            bigip.headers.update({"X-F5-REST-Coordination-Id": None})

    return bigip


def _load_response(response):
    """
    Load the response from json data, return the dictionary or raw text
    """

    try:
        data = salt.utils.json.loads(response.text)
    except ValueError:
        data = response.text

    ret = {"code": response.status_code, "content": data}

    return ret


def _load_connection_error(hostname, error):
    """
    Format and Return a connection error
    """

    ret = {
        "code": None,
        "content": (
            "Error: Unable to connect to the bigip device: {host}\n{error}".format(
                host=hostname, error=error
            )
        ),
    }

    return ret


def _loop_payload(params):
    """
    Pass in a dictionary of parameters, loop through them and build a payload containing,
    parameters who's values are not None.
    """

    # construct the payload
    payload = {}

    # set the payload
    for param, value in params.items():
        if value is not None:
            payload[param] = value

    return payload


def _build_list(option_value, item_kind):
    """
    pass in an option to check for a list of items, create a list of dictionary of items to set
    for this option
    """
    # specify profiles if provided
    if option_value is not None:

        items = []

        # if user specified none, return an empty list
        if option_value == "none":
            return items

        # was a list already passed in?
        if not isinstance(option_value, list):
            values = option_value.split(",")
        else:
            values = option_value

        for value in values:
            # sometimes the bigip just likes a plain ol list of items
            if item_kind is None:
                items.append(value)
            # other times it's picky and likes key value pairs...
            else:
                items.append({"kind": item_kind, "name": value})
        return items
    return None


def _determine_toggles(payload, toggles):
    """
    BigIP can't make up its mind if it likes yes / no or true or false.
    Figure out what it likes to hear without confusing the user.
    """

    for toggle, definition in toggles.items():
        # did the user specify anything?
        if definition["value"] is not None:
            # test for yes_no toggle
            if (
                definition["value"] is True or definition["value"] == "yes"
            ) and definition["type"] == "yes_no":
                payload[toggle] = "yes"
            elif (
                definition["value"] is False or definition["value"] == "no"
            ) and definition["type"] == "yes_no":
                payload[toggle] = "no"

            # test for true_false toggle
            if (
                definition["value"] is True or definition["value"] == "yes"
            ) and definition["type"] == "true_false":
                payload[toggle] = True
            elif (
                definition["value"] is False or definition["value"] == "no"
            ) and definition["type"] == "true_false":
                payload[toggle] = False

    return payload


def _set_value(value):
    """
    A function to detect if user is trying to pass a dictionary or list.  parse it and return a
    dictionary list or a string
    """
    # don't continue if already an acceptable data-type
    if isinstance(value, bool) or isinstance(value, dict) or isinstance(value, list):
        return value

    # check if json
    if value.startswith("j{") and value.endswith("}j"):

        value = value.replace("j{", "{")
        value = value.replace("}j", "}")

        try:
            return salt.utils.json.loads(value)
        except Exception:  # pylint: disable=broad-except
            raise salt.exceptions.CommandExecutionError

    # detect list of dictionaries
    if "|" in value and r"\|" not in value:
        values = value.split("|")
        items = []
        for value in values:
            items.append(_set_value(value))
        return items

    # parse out dictionary if detected
    if ":" in value and r"\:" not in value:
        options = {}
        # split out pairs
        key_pairs = value.split(",")
        for key_pair in key_pairs:
            k = key_pair.split(":")[0]
            v = key_pair.split(":")[1]
            options[k] = v
        return options

    # try making a list
    elif "," in value and r"\," not in value:
        value_items = value.split(",")
        return value_items

    # just return a string
    else:

        # remove escape chars if added
        if r"\|" in value:
            value = value.replace(r"\|", "|")

        if r"\:" in value:
            value = value.replace(r"\:", ":")

        if r"\," in value:
            value = value.replace(r"\,", ",")

        return value


def start_transaction(hostname, username, password, label):
    """
    A function to connect to a bigip device and start a new transaction.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    label
        The name / alias for this transaction.  The actual transaction
        id will be stored within a grain called ``bigip_f5_trans:<label>``

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.start_transaction bigip admin admin my_transaction

    """

    # build the session
    bigip_session = _build_session(username, password)

    payload = {}

    # post to REST to get trans id
    try:
        response = bigip_session.post(
            BIG_IP_URL_BASE.format(host=hostname) + "/transaction",
            data=salt.utils.json.dumps(payload),
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    # extract the trans_id
    data = _load_response(response)

    if data["code"] == 200:

        trans_id = data["content"]["transId"]

        __salt__["grains.setval"]("bigip_f5_trans", {label: trans_id})

        return (
            "Transaction: {trans_id} - has successfully been stored in the grain:"
            " bigip_f5_trans:{label}".format(trans_id=trans_id, label=label)
        )
    else:
        return data


def list_transaction(hostname, username, password, label):
    """
    A function to connect to a bigip device and list an existing transaction.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    label
        the label of this transaction stored within the grain:
        ``bigip_f5_trans:<label>``

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.list_transaction bigip admin admin my_transaction

    """

    # build the session
    bigip_session = _build_session(username, password)

    # pull the trans id from the grain
    trans_id = __salt__["grains.get"](f"bigip_f5_trans:{label}")

    if trans_id:

        # post to REST to get trans id
        try:
            response = bigip_session.get(
                BIG_IP_URL_BASE.format(host=hostname)
                + f"/transaction/{trans_id}/commands"
            )
            return _load_response(response)
        except requests.exceptions.ConnectionError as e:
            return _load_connection_error(hostname, e)
    else:
        return (
            "Error: the label for this transaction was not defined as a grain.  Begin a"
            " new transaction using the bigip.start_transaction function"
        )


def commit_transaction(hostname, username, password, label):
    """
    A function to connect to a bigip device and commit an existing transaction.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    label
        the label of this transaction stored within the grain:
        ``bigip_f5_trans:<label>``

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.commit_transaction bigip admin admin my_transaction
    """

    # build the session
    bigip_session = _build_session(username, password)

    # pull the trans id from the grain
    trans_id = __salt__["grains.get"](f"bigip_f5_trans:{label}")

    if trans_id:

        payload = {}
        payload["state"] = "VALIDATING"

        # patch to REST to get trans id
        try:
            response = bigip_session.patch(
                BIG_IP_URL_BASE.format(host=hostname) + f"/transaction/{trans_id}",
                data=salt.utils.json.dumps(payload),
            )
            return _load_response(response)
        except requests.exceptions.ConnectionError as e:
            return _load_connection_error(hostname, e)
    else:
        return (
            "Error: the label for this transaction was not defined as a grain.  Begin a"
            " new transaction using the bigip.start_transaction function"
        )


def delete_transaction(hostname, username, password, label):
    """
    A function to connect to a bigip device and delete an existing transaction.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    label
        The label of this transaction stored within the grain:
        ``bigip_f5_trans:<label>``

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.delete_transaction bigip admin admin my_transaction
    """

    # build the session
    bigip_session = _build_session(username, password)

    # pull the trans id from the grain
    trans_id = __salt__["grains.get"](f"bigip_f5_trans:{label}")

    if trans_id:

        # patch to REST to get trans id
        try:
            response = bigip_session.delete(
                BIG_IP_URL_BASE.format(host=hostname) + f"/transaction/{trans_id}"
            )
            return _load_response(response)
        except requests.exceptions.ConnectionError as e:
            return _load_connection_error(hostname, e)
    else:
        return (
            "Error: the label for this transaction was not defined as a grain.  Begin a"
            " new transaction using the bigip.start_transaction function"
        )


def list_node(hostname, username, password, name=None, trans_label=None):
    """
    A function to connect to a bigip device and list all nodes or a specific node.


    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the node to list. If no name is specified than all nodes
        will be listed.
    trans_label
        The label of the transaction stored within the grain:
        ``bigip_f5_trans:<label>``

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.list_node bigip admin admin my-node
    """

    # build sessions
    bigip_session = _build_session(username, password, trans_label)

    # get to REST
    try:
        if name:
            response = bigip_session.get(
                BIG_IP_URL_BASE.format(host=hostname) + f"/ltm/node/{name}"
            )
        else:
            response = bigip_session.get(
                BIG_IP_URL_BASE.format(host=hostname) + "/ltm/node"
            )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def create_node(hostname, username, password, name, address, trans_label=None):
    """
    A function to connect to a bigip device and create a node.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the node
    address
        The address of the node
    trans_label
        The label of the transaction stored within the grain:
        ``bigip_f5_trans:<label>``

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.create_node bigip admin admin 10.1.1.2
    """

    # build session
    bigip_session = _build_session(username, password, trans_label)

    # construct the payload
    payload = {}
    payload["name"] = name
    payload["address"] = address

    # post to REST
    try:
        response = bigip_session.post(
            BIG_IP_URL_BASE.format(host=hostname) + "/ltm/node",
            data=salt.utils.json.dumps(payload),
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def modify_node(
    hostname,
    username,
    password,
    name,
    connection_limit=None,
    description=None,
    dynamic_ratio=None,
    logging=None,
    monitor=None,
    rate_limit=None,
    ratio=None,
    session=None,
    state=None,
    trans_label=None,
):
    """
    A function to connect to a bigip device and modify an existing node.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the node to modify
    connection_limit
        [integer]
    description
        [string]
    dynamic_ratio
        [integer]
    logging
        [enabled | disabled]
    monitor
        [[name] | none | default]
    rate_limit
        [integer]
    ratio
        [integer]
    session
        [user-enabled | user-disabled]
    state
        [user-down | user-up ]
    trans_label
        The label of the transaction stored within the grain:
        ``bigip_f5_trans:<label>``

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.modify_node bigip admin admin 10.1.1.2 ratio=2 logging=enabled
    """

    params = {
        "connection-limit": connection_limit,
        "description": description,
        "dynamic-ratio": dynamic_ratio,
        "logging": logging,
        "monitor": monitor,
        "rate-limit": rate_limit,
        "ratio": ratio,
        "session": session,
        "state": state,
    }

    # build session
    bigip_session = _build_session(username, password, trans_label)

    # build payload
    payload = _loop_payload(params)
    payload["name"] = name

    # put to REST
    try:
        response = bigip_session.put(
            BIG_IP_URL_BASE.format(host=hostname) + f"/ltm/node/{name}",
            data=salt.utils.json.dumps(payload),
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def delete_node(hostname, username, password, name, trans_label=None):
    """
    A function to connect to a bigip device and delete a specific node.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the node which will be deleted.
    trans_label
        The label of the transaction stored within the grain:
        ``bigip_f5_trans:<label>``

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.delete_node bigip admin admin my-node
    """

    # build session
    bigip_session = _build_session(username, password, trans_label)

    # delete to REST
    try:
        response = bigip_session.delete(
            BIG_IP_URL_BASE.format(host=hostname) + f"/ltm/node/{name}"
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    if _load_response(response) == "":
        return True
    else:
        return _load_response(response)


def list_pool(hostname, username, password, name=None):
    """
    A function to connect to a bigip device and list all pools or a specific pool.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to list. If no name is specified then all pools
        will be listed.

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.list_pool bigip admin admin my-pool
    """

    # build sessions
    bigip_session = _build_session(username, password)

    # get to REST
    try:
        if name:
            response = bigip_session.get(
                BIG_IP_URL_BASE.format(host=hostname)
                + f"/ltm/pool/{name}/?expandSubcollections=true"
            )
        else:
            response = bigip_session.get(
                BIG_IP_URL_BASE.format(host=hostname) + "/ltm/pool"
            )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def create_pool(
    hostname,
    username,
    password,
    name,
    members=None,
    allow_nat=None,
    allow_snat=None,
    description=None,
    gateway_failsafe_device=None,
    ignore_persisted_weight=None,
    ip_tos_to_client=None,
    ip_tos_to_server=None,
    link_qos_to_client=None,
    link_qos_to_server=None,
    load_balancing_mode=None,
    min_active_members=None,
    min_up_members=None,
    min_up_members_action=None,
    min_up_members_checking=None,
    monitor=None,
    profiles=None,
    queue_depth_limit=None,
    queue_on_connection_limit=None,
    queue_time_limit=None,
    reselect_tries=None,
    service_down_action=None,
    slow_ramp_time=None,
):
    """
    A function to connect to a bigip device and create a pool.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to create.
    members
        List of comma delimited pool members to add to the pool.
        i.e. 10.1.1.1:80,10.1.1.2:80,10.1.1.3:80
    allow_nat
        [yes | no]
    allow_snat
        [yes | no]
    description
        [string]
    gateway_failsafe_device
        [string]
    ignore_persisted_weight
        [enabled | disabled]
    ip_tos_to_client
        [pass-through | [integer]]
    ip_tos_to_server
        [pass-through | [integer]]
    link_qos_to_client
        [pass-through | [integer]]
    link_qos_to_server
        [pass-through | [integer]]
    load_balancing_mode
        [dynamic-ratio-member | dynamic-ratio-node |
        fastest-app-response | fastest-node |
        least-connections-members |
        least-connections-node |
        least-sessions |
        observed-member | observed-node |
        predictive-member | predictive-node |
        ratio-least-connections-member |
        ratio-least-connections-node |
        ratio-member | ratio-node | ratio-session |
        round-robin | weighted-least-connections-member |
        weighted-least-connections-node]
    min_active_members
        [integer]
    min_up_members
        [integer]
    min_up_members_action
        [failover | reboot | restart-all]
    min_up_members_checking
        [enabled | disabled]
    monitor
        [name]
    profiles
        [none | profile_name]
    queue_depth_limit
        [integer]
    queue_on_connection_limit
        [enabled | disabled]
    queue_time_limit
        [integer]
    reselect_tries
        [integer]
    service_down_action
        [drop | none | reselect | reset]
    slow_ramp_time
        [integer]

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.create_pool bigip admin admin my-pool 10.1.1.1:80,10.1.1.2:80,10.1.1.3:80 monitor=http
    """

    params = {
        "description": description,
        "gateway-failsafe-device": gateway_failsafe_device,
        "ignore-persisted-weight": ignore_persisted_weight,
        "ip-tos-to-client": ip_tos_to_client,
        "ip-tos-to-server": ip_tos_to_server,
        "link-qos-to-client": link_qos_to_client,
        "link-qos-to-server": link_qos_to_server,
        "load-balancing-mode": load_balancing_mode,
        "min-active-members": min_active_members,
        "min-up-members": min_up_members,
        "min-up-members-action": min_up_members_action,
        "min-up-members-checking": min_up_members_checking,
        "monitor": monitor,
        "profiles": profiles,
        "queue-on-connection-limit": queue_on_connection_limit,
        "queue-depth-limit": queue_depth_limit,
        "queue-time-limit": queue_time_limit,
        "reselect-tries": reselect_tries,
        "service-down-action": service_down_action,
        "slow-ramp-time": slow_ramp_time,
    }

    # some options take yes no others take true false.  Figure out when to use which without
    # confusing the end user
    toggles = {
        "allow-nat": {"type": "yes_no", "value": allow_nat},
        "allow-snat": {"type": "yes_no", "value": allow_snat},
    }

    # build payload
    payload = _loop_payload(params)
    payload["name"] = name

    # determine toggles
    payload = _determine_toggles(payload, toggles)

    # specify members if provided
    if members is not None:
        payload["members"] = _build_list(members, "ltm:pool:members")

    # build session
    bigip_session = _build_session(username, password)

    # post to REST
    try:
        response = bigip_session.post(
            BIG_IP_URL_BASE.format(host=hostname) + "/ltm/pool",
            data=salt.utils.json.dumps(payload),
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def modify_pool(
    hostname,
    username,
    password,
    name,
    allow_nat=None,
    allow_snat=None,
    description=None,
    gateway_failsafe_device=None,
    ignore_persisted_weight=None,
    ip_tos_to_client=None,
    ip_tos_to_server=None,
    link_qos_to_client=None,
    link_qos_to_server=None,
    load_balancing_mode=None,
    min_active_members=None,
    min_up_members=None,
    min_up_members_action=None,
    min_up_members_checking=None,
    monitor=None,
    profiles=None,
    queue_depth_limit=None,
    queue_on_connection_limit=None,
    queue_time_limit=None,
    reselect_tries=None,
    service_down_action=None,
    slow_ramp_time=None,
):
    """
    A function to connect to a bigip device and modify an existing pool.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to modify.
    allow_nat
        [yes | no]
    allow_snat
        [yes | no]
    description
        [string]
    gateway_failsafe_device
        [string]
    ignore_persisted_weight
        [yes | no]
    ip_tos_to_client
        [pass-through | [integer]]
    ip_tos_to_server
        [pass-through | [integer]]
    link_qos_to_client
        [pass-through | [integer]]
    link_qos_to_server
        [pass-through | [integer]]
    load_balancing_mode
        [dynamic-ratio-member | dynamic-ratio-node |
        fastest-app-response | fastest-node |
        least-connections-members |
        least-connections-node |
        least-sessions |
        observed-member | observed-node |
        predictive-member | predictive-node |
        ratio-least-connections-member |
        ratio-least-connections-node |
        ratio-member | ratio-node | ratio-session |
        round-robin | weighted-least-connections-member |
        weighted-least-connections-node]
    min_active_members
        [integer]
    min_up_members
        [integer]
    min_up_members_action
        [failover | reboot | restart-all]
    min_up_members_checking
        [enabled | disabled]
    monitor
        [name]
    profiles
        [none | profile_name]
    queue_on_connection_limit
        [enabled | disabled]
    queue_depth_limit
        [integer]
    queue_time_limit
        [integer]
    reselect_tries
        [integer]
    service_down_action
        [drop | none | reselect | reset]
    slow_ramp_time
        [integer]

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.modify_pool bigip admin admin my-pool 10.1.1.1:80,10.1.1.2:80,10.1.1.3:80 min_active_members=1
    """

    params = {
        "description": description,
        "gateway-failsafe-device": gateway_failsafe_device,
        "ignore-persisted-weight": ignore_persisted_weight,
        "ip-tos-to-client": ip_tos_to_client,
        "ip-tos-to-server": ip_tos_to_server,
        "link-qos-to-client": link_qos_to_client,
        "link-qos-to-server": link_qos_to_server,
        "load-balancing-mode": load_balancing_mode,
        "min-active-members": min_active_members,
        "min-up-members": min_up_members,
        "min-up_members-action": min_up_members_action,
        "min-up-members-checking": min_up_members_checking,
        "monitor": monitor,
        "profiles": profiles,
        "queue-on-connection-limit": queue_on_connection_limit,
        "queue-depth-limit": queue_depth_limit,
        "queue-time-limit": queue_time_limit,
        "reselect-tries": reselect_tries,
        "service-down-action": service_down_action,
        "slow-ramp-time": slow_ramp_time,
    }

    # some options take yes no others take true false.  Figure out when to use which without
    # confusing the end user
    toggles = {
        "allow-nat": {"type": "yes_no", "value": allow_nat},
        "allow-snat": {"type": "yes_no", "value": allow_snat},
    }

    # build payload
    payload = _loop_payload(params)
    payload["name"] = name

    # determine toggles
    payload = _determine_toggles(payload, toggles)

    # build session
    bigip_session = _build_session(username, password)

    # post to REST
    try:
        response = bigip_session.put(
            BIG_IP_URL_BASE.format(host=hostname) + f"/ltm/pool/{name}",
            data=salt.utils.json.dumps(payload),
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def delete_pool(hostname, username, password, name):
    """
    A function to connect to a bigip device and delete a specific pool.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool which will be deleted

    CLI Example

    .. code-block:: bash

        salt '*' bigip.delete_node bigip admin admin my-pool
    """

    # build session
    bigip_session = _build_session(username, password)

    # delete to REST
    try:
        response = bigip_session.delete(
            BIG_IP_URL_BASE.format(host=hostname) + f"/ltm/pool/{name}"
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    if _load_response(response) == "":
        return True
    else:
        return _load_response(response)


def replace_pool_members(hostname, username, password, name, members):
    """
    A function to connect to a bigip device and replace members of an existing pool with new members.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to modify
    members
        List of comma delimited pool members to replace existing members with.
        i.e. 10.1.1.1:80,10.1.1.2:80,10.1.1.3:80

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.replace_pool_members bigip admin admin my-pool 10.2.2.1:80,10.2.2.2:80,10.2.2.3:80
    """

    payload = {}
    payload["name"] = name
    # specify members if provided
    if members is not None:

        if isinstance(members, str):
            members = members.split(",")

        pool_members = []
        for member in members:

            # check to see if already a dictionary ( for states)
            if isinstance(member, dict):

                # check for state alternative name 'member_state', replace with state
                if "member_state" in member.keys():
                    member["state"] = member.pop("member_state")

                # replace underscore with dash
                for key in member:
                    new_key = key.replace("_", "-")
                    member[new_key] = member.pop(key)

                pool_members.append(member)

            # parse string passed via execution command (for executions)
            else:
                pool_members.append({"name": member, "address": member.split(":")[0]})

        payload["members"] = pool_members

    # build session
    bigip_session = _build_session(username, password)

    # put to REST
    try:
        response = bigip_session.put(
            BIG_IP_URL_BASE.format(host=hostname) + f"/ltm/pool/{name}",
            data=salt.utils.json.dumps(payload),
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def add_pool_member(hostname, username, password, name, member):
    """
    A function to connect to a bigip device and add a new member to an existing pool.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to modify
    member
        The name of the member to add
        i.e. 10.1.1.2:80

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.add_pool_members bigip admin admin my-pool 10.2.2.1:80
    """

    # for states
    if isinstance(member, dict):

        # check for state alternative name 'member_state', replace with state
        if "member_state" in member.keys():
            member["state"] = member.pop("member_state")

        # replace underscore with dash
        for key in member:
            new_key = key.replace("_", "-")
            member[new_key] = member.pop(key)

        payload = member
    # for execution
    else:

        payload = {"name": member, "address": member.split(":")[0]}

    # build session
    bigip_session = _build_session(username, password)

    # post to REST
    try:
        response = bigip_session.post(
            BIG_IP_URL_BASE.format(host=hostname) + f"/ltm/pool/{name}/members",
            data=salt.utils.json.dumps(payload),
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def modify_pool_member(
    hostname,
    username,
    password,
    name,
    member,
    connection_limit=None,
    description=None,
    dynamic_ratio=None,
    inherit_profile=None,
    logging=None,
    monitor=None,
    priority_group=None,
    profiles=None,
    rate_limit=None,
    ratio=None,
    session=None,
    state=None,
):
    """
    A function to connect to a bigip device and modify an existing member of a pool.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to modify
    member
        The name of the member to modify i.e. 10.1.1.2:80
    connection_limit
        [integer]
    description
        [string]
    dynamic_ratio
        [integer]
    inherit_profile
        [enabled | disabled]
    logging
        [enabled | disabled]
    monitor
        [name]
    priority_group
        [integer]
    profiles
        [none | profile_name]
    rate_limit
        [integer]
    ratio
        [integer]
    session
        [user-enabled | user-disabled]
    state
        [ user-up | user-down ]

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.modify_pool_member bigip admin admin my-pool 10.2.2.1:80 state=use-down session=user-disabled
    """

    params = {
        "connection-limit": connection_limit,
        "description": description,
        "dynamic-ratio": dynamic_ratio,
        "inherit-profile": inherit_profile,
        "logging": logging,
        "monitor": monitor,
        "priority-group": priority_group,
        "profiles": profiles,
        "rate-limit": rate_limit,
        "ratio": ratio,
        "session": session,
        "state": state,
    }

    # build session
    bigip_session = _build_session(username, password)

    # build payload
    payload = _loop_payload(params)

    # put to REST
    try:
        response = bigip_session.put(
            BIG_IP_URL_BASE.format(host=hostname)
            + f"/ltm/pool/{name}/members/{member}",
            data=salt.utils.json.dumps(payload),
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def delete_pool_member(hostname, username, password, name, member):
    """
    A function to connect to a bigip device and delete a specific pool.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to modify
    member
        The name of the pool member to delete

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.delete_pool_member bigip admin admin my-pool 10.2.2.2:80
    """

    # build session
    bigip_session = _build_session(username, password)

    # delete to REST
    try:
        response = bigip_session.delete(
            BIG_IP_URL_BASE.format(host=hostname) + f"/ltm/pool/{name}/members/{member}"
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    if _load_response(response) == "":
        return True
    else:
        return _load_response(response)


def list_virtual(hostname, username, password, name=None):
    """
    A function to connect to a bigip device and list all virtuals or a specific virtual.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the virtual to list. If no name is specified than all
        virtuals will be listed.

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.list_virtual bigip admin admin my-virtual
    """

    # build sessions
    bigip_session = _build_session(username, password)

    # get to REST
    try:
        if name:
            response = bigip_session.get(
                BIG_IP_URL_BASE.format(host=hostname)
                + f"/ltm/virtual/{name}/?expandSubcollections=true"
            )
        else:
            response = bigip_session.get(
                BIG_IP_URL_BASE.format(host=hostname) + "/ltm/virtual"
            )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def create_virtual(
    hostname,
    username,
    password,
    name,
    destination,
    pool=None,
    address_status=None,
    auto_lasthop=None,
    bwc_policy=None,
    cmp_enabled=None,
    connection_limit=None,
    dhcp_relay=None,
    description=None,
    fallback_persistence=None,
    flow_eviction_policy=None,
    gtm_score=None,
    ip_forward=None,
    ip_protocol=None,
    internal=None,
    twelve_forward=None,
    last_hop_pool=None,
    mask=None,
    mirror=None,
    nat64=None,
    persist=None,
    profiles=None,
    policies=None,
    rate_class=None,
    rate_limit=None,
    rate_limit_mode=None,
    rate_limit_dst=None,
    rate_limit_src=None,
    rules=None,
    related_rules=None,
    reject=None,
    source=None,
    source_address_translation=None,
    source_port=None,
    state=None,
    traffic_classes=None,
    translate_address=None,
    translate_port=None,
    vlans=None,
):
    r"""
    A function to connect to a bigip device and create a virtual server.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the virtual to create
    destination
        [ [virtual_address_name:port] | [ipv4:port] | [ipv6.port] ]
    pool
        [ [pool_name] | none]
    address_status
        [yes | no]
    auto_lasthop
        [default | enabled | disabled ]
    bwc_policy
        [none] | string]
    cmp_enabled
        [yes | no]
    dhcp_relay
        [yes | no]
    connection_limit
        [integer]
    description
        [string]
    state
        [disabled | enabled]
    fallback_persistence
        [none | [profile name] ]
    flow_eviction_policy
        [none | [eviction policy name] ]
    gtm_score
        [integer]
    ip_forward
        [yes | no]
    ip_protocol
        [any | protocol]
    internal
        [yes | no]
    twelve_forward
        (12-forward)
        [yes | no]
    last_hop-pool
        [ [pool_name] | none]
    mask
        { [ipv4] | [ipv6] }
    mirror
        { [disabled | enabled | none] }
    nat64
        [enabled | disabled]
    persist
        [none | profile1,profile2,profile3 ... ]
    profiles
        [none | default | profile1,profile2,profile3 ... ]
    policies
        [none | default | policy1,policy2,policy3 ... ]
    rate_class
        [name]
    rate_limit
        [integer]
    rate_limit_mode
        [destination | object | object-destination |
        object-source | object-source-destination |
        source | source-destination]
    rate_limit_dst
        [integer]
    rate_limitçsrc
        [integer]
    rules
        [none | [rule_one,rule_two ...] ]
    related_rules
        [none | [rule_one,rule_two ...] ]
    reject
        [yes | no]
    source
        { [ipv4[/prefixlen]] | [ipv6[/prefixlen]] }
    source_address_translation
        [none | snat:pool_name | lsn | automap ]
    source_port
        [change | preserve | preserve-strict]
    state
        [enabled | disabled]
    traffic_classes
        [none | default | class_one,class_two ... ]
    translate_address
        [enabled | disabled]
    translate_port
        [enabled | disabled]
    vlans
        [none | default | [enabled|disabled]:vlan1,vlan2,vlan3 ... ]

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.create_virtual bigip admin admin my-virtual-3 26.2.2.5:80 \
            pool=my-http-pool-http profiles=http,tcp

        salt '*' bigip.create_virtual bigip admin admin my-virtual-3 43.2.2.5:80 \
            pool=test-http-pool-http profiles=http,websecurity persist=cookie,hash \
            policies=asm_auto_l7_policy__http-virtual \
            rules=_sys_APM_ExchangeSupport_helper,_sys_https_redirect \
            related_rules=_sys_APM_activesync,_sys_APM_ExchangeSupport_helper \
            source_address_translation=snat:my-snat-pool \
            translate_address=enabled translate_port=enabled \
            traffic_classes=my-class,other-class \
            vlans=enabled:external,internal

    """

    params = {
        "pool": pool,
        "auto-lasthop": auto_lasthop,
        "bwc-policy": bwc_policy,
        "connection-limit": connection_limit,
        "description": description,
        "fallback-persistence": fallback_persistence,
        "flow-eviction-policy": flow_eviction_policy,
        "gtm-score": gtm_score,
        "ip-protocol": ip_protocol,
        "last-hop-pool": last_hop_pool,
        "mask": mask,
        "mirror": mirror,
        "nat64": nat64,
        "persist": persist,
        "rate-class": rate_class,
        "rate-limit": rate_limit,
        "rate-limit-mode": rate_limit_mode,
        "rate-limit-dst": rate_limit_dst,
        "rate-limit-src": rate_limit_src,
        "source": source,
        "source-port": source_port,
        "translate-address": translate_address,
        "translate-port": translate_port,
    }

    # some options take yes no others take true false.  Figure out when to use which without
    # confusing the end user
    toggles = {
        "address-status": {"type": "yes_no", "value": address_status},
        "cmp-enabled": {"type": "yes_no", "value": cmp_enabled},
        "dhcp-relay": {"type": "true_false", "value": dhcp_relay},
        "reject": {"type": "true_false", "value": reject},
        "12-forward": {"type": "true_false", "value": twelve_forward},
        "internal": {"type": "true_false", "value": internal},
        "ip-forward": {"type": "true_false", "value": ip_forward},
    }

    # build session
    bigip_session = _build_session(username, password)

    # build payload
    payload = _loop_payload(params)

    payload["name"] = name
    payload["destination"] = destination

    # determine toggles
    payload = _determine_toggles(payload, toggles)

    # specify profiles if provided
    if profiles is not None:
        payload["profiles"] = _build_list(profiles, "ltm:virtual:profile")

    # specify persist if provided
    if persist is not None:
        payload["persist"] = _build_list(persist, "ltm:virtual:persist")

    # specify policies if provided
    if policies is not None:
        payload["policies"] = _build_list(policies, "ltm:virtual:policy")

    # specify rules if provided
    if rules is not None:
        payload["rules"] = _build_list(rules, None)

    # specify related-rules if provided
    if related_rules is not None:
        payload["related-rules"] = _build_list(related_rules, None)

    # handle source-address-translation
    if source_address_translation is not None:

        # check to see if this is already a dictionary first
        if isinstance(source_address_translation, dict):
            payload["source-address-translation"] = source_address_translation
        elif source_address_translation == "none":
            payload["source-address-translation"] = {"pool": "none", "type": "none"}
        elif source_address_translation == "automap":
            payload["source-address-translation"] = {"pool": "none", "type": "automap"}
        elif source_address_translation == "lsn":
            payload["source-address-translation"] = {"pool": "none", "type": "lsn"}
        elif source_address_translation.startswith("snat"):
            snat_pool = source_address_translation.split(":")[1]
            payload["source-address-translation"] = {"pool": snat_pool, "type": "snat"}

    # specify related-rules if provided
    if traffic_classes is not None:
        payload["traffic-classes"] = _build_list(traffic_classes, None)

    # handle vlans
    if vlans is not None:
        # ceck to see if vlans is a dictionary (used when state makes use of function)
        if isinstance(vlans, dict):
            try:
                payload["vlans"] = vlans["vlan_ids"]
                if vlans["enabled"]:
                    payload["vlans-enabled"] = True
                elif vlans["disabled"]:
                    payload["vlans-disabled"] = True
            except Exception:  # pylint: disable=broad-except
                return (
                    "Error: Unable to Parse vlans dictionary: \n\tvlans={vlans}".format(
                        vlans=vlans
                    )
                )
        elif vlans == "none":
            payload["vlans"] = "none"
        elif vlans == "default":
            payload["vlans"] = "default"
        elif isinstance(vlans, str) and (
            vlans.startswith("enabled") or vlans.startswith("disabled")
        ):
            try:
                vlans_setting = vlans.split(":")[0]
                payload["vlans"] = vlans.split(":")[1].split(",")
                if vlans_setting == "disabled":
                    payload["vlans-disabled"] = True
                elif vlans_setting == "enabled":
                    payload["vlans-enabled"] = True
            except Exception:  # pylint: disable=broad-except
                return "Error: Unable to Parse vlans option: \n\tvlans={vlans}".format(
                    vlans=vlans
                )
        else:
            return "Error: vlans must be a dictionary or string."

    # determine state
    if state is not None:
        if state == "enabled":
            payload["enabled"] = True
        elif state == "disabled":
            payload["disabled"] = True

    # post to REST
    try:
        response = bigip_session.post(
            BIG_IP_URL_BASE.format(host=hostname) + "/ltm/virtual",
            data=salt.utils.json.dumps(payload),
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def modify_virtual(
    hostname,
    username,
    password,
    name,
    destination=None,
    pool=None,
    address_status=None,
    auto_lasthop=None,
    bwc_policy=None,
    cmp_enabled=None,
    connection_limit=None,
    dhcp_relay=None,
    description=None,
    fallback_persistence=None,
    flow_eviction_policy=None,
    gtm_score=None,
    ip_forward=None,
    ip_protocol=None,
    internal=None,
    twelve_forward=None,
    last_hop_pool=None,
    mask=None,
    mirror=None,
    nat64=None,
    persist=None,
    profiles=None,
    policies=None,
    rate_class=None,
    rate_limit=None,
    rate_limit_mode=None,
    rate_limit_dst=None,
    rate_limit_src=None,
    rules=None,
    related_rules=None,
    reject=None,
    source=None,
    source_address_translation=None,
    source_port=None,
    state=None,
    traffic_classes=None,
    translate_address=None,
    translate_port=None,
    vlans=None,
):
    """
    A function to connect to a bigip device and modify an existing virtual server.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the virtual to modify
    destination
        [ [virtual_address_name:port] | [ipv4:port] | [ipv6.port] ]
    pool
        [ [pool_name] | none]
    address_status
        [yes | no]
    auto_lasthop
        [default | enabled | disabled ]
    bwc_policy
        [none] | string]
    cmp_enabled
        [yes | no]
    dhcp_relay
        [yes | no}
    connection_limit
        [integer]
    description
        [string]
    state
        [disabled | enabled]
    fallback_persistence
        [none | [profile name] ]
    flow_eviction_policy
        [none | [eviction policy name] ]
    gtm_score
        [integer]
    ip_forward
        [yes | no]
    ip_protocol
        [any | protocol]
    internal
        [yes | no]
    twelve_forward
        (12-forward)
        [yes | no]
    last_hop-pool
        [ [pool_name] | none]
    mask
        { [ipv4] | [ipv6] }
    mirror
        { [disabled | enabled | none] }
    nat64
        [enabled | disabled]
    persist
        [none | profile1,profile2,profile3 ... ]
    profiles
        [none | default | profile1,profile2,profile3 ... ]
    policies
        [none | default | policy1,policy2,policy3 ... ]
    rate_class
        [name]
    rate_limit
        [integer]
    rate_limitr_mode
        [destination | object | object-destination |
        object-source | object-source-destination |
        source | source-destination]
    rate_limit_dst
        [integer]
    rate_limit_src
        [integer]
    rules
        [none | [rule_one,rule_two ...] ]
    related_rules
        [none | [rule_one,rule_two ...] ]
    reject
        [yes | no]
    source
        { [ipv4[/prefixlen]] | [ipv6[/prefixlen]] }
    source_address_translation
        [none | snat:pool_name | lsn | automap ]
    source_port
        [change | preserve | preserve-strict]
    state
        [enabled | disable]
    traffic_classes
        [none | default | class_one,class_two ... ]
    translate_address
        [enabled | disabled]
    translate_port
        [enabled | disabled]
    vlans
        [none | default | [enabled|disabled]:vlan1,vlan2,vlan3 ... ]

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.modify_virtual bigip admin admin my-virtual source_address_translation=none
        salt '*' bigip.modify_virtual bigip admin admin my-virtual rules=my-rule,my-other-rule
    """

    params = {
        "destination": destination,
        "pool": pool,
        "auto-lasthop": auto_lasthop,
        "bwc-policy": bwc_policy,
        "connection-limit": connection_limit,
        "description": description,
        "fallback-persistence": fallback_persistence,
        "flow-eviction-policy": flow_eviction_policy,
        "gtm-score": gtm_score,
        "ip-protocol": ip_protocol,
        "last-hop-pool": last_hop_pool,
        "mask": mask,
        "mirror": mirror,
        "nat64": nat64,
        "persist": persist,
        "rate-class": rate_class,
        "rate-limit": rate_limit,
        "rate-limit-mode": rate_limit_mode,
        "rate-limit-dst": rate_limit_dst,
        "rate-limit-src": rate_limit_src,
        "source": source,
        "source-port": source_port,
        "translate-address": translate_address,
        "translate-port": translate_port,
    }

    # some options take yes no others take true false.  Figure out when to use which without
    # confusing the end user
    toggles = {
        "address-status": {"type": "yes_no", "value": address_status},
        "cmp-enabled": {"type": "yes_no", "value": cmp_enabled},
        "dhcp-relay": {"type": "true_false", "value": dhcp_relay},
        "reject": {"type": "true_false", "value": reject},
        "12-forward": {"type": "true_false", "value": twelve_forward},
        "internal": {"type": "true_false", "value": internal},
        "ip-forward": {"type": "true_false", "value": ip_forward},
    }

    # build session
    bigip_session = _build_session(username, password)

    # build payload
    payload = _loop_payload(params)
    payload["name"] = name

    # determine toggles
    payload = _determine_toggles(payload, toggles)

    # specify profiles if provided
    if profiles is not None:
        payload["profiles"] = _build_list(profiles, "ltm:virtual:profile")

    # specify persist if provided
    if persist is not None:
        payload["persist"] = _build_list(persist, "ltm:virtual:persist")

    # specify policies if provided
    if policies is not None:
        payload["policies"] = _build_list(policies, "ltm:virtual:policy")

    # specify rules if provided
    if rules is not None:
        payload["rules"] = _build_list(rules, None)

    # specify related-rules if provided
    if related_rules is not None:
        payload["related-rules"] = _build_list(related_rules, None)

    # handle source-address-translation
    if source_address_translation is not None:
        if source_address_translation == "none":
            payload["source-address-translation"] = {"pool": "none", "type": "none"}
        elif source_address_translation == "automap":
            payload["source-address-translation"] = {"pool": "none", "type": "automap"}
        elif source_address_translation == "lsn":
            payload["source-address-translation"] = {"pool": "none", "type": "lsn"}
        elif source_address_translation.startswith("snat"):
            snat_pool = source_address_translation.split(":")[1]
            payload["source-address-translation"] = {"pool": snat_pool, "type": "snat"}

    # specify related-rules if provided
    if traffic_classes is not None:
        payload["traffic-classes"] = _build_list(traffic_classes, None)

    # handle vlans
    if vlans is not None:
        # ceck to see if vlans is a dictionary (used when state makes use of function)
        if isinstance(vlans, dict):
            try:
                payload["vlans"] = vlans["vlan_ids"]
                if vlans["enabled"]:
                    payload["vlans-enabled"] = True
                elif vlans["disabled"]:
                    payload["vlans-disabled"] = True
            except Exception:  # pylint: disable=broad-except
                return (
                    "Error: Unable to Parse vlans dictionary: \n\tvlans={vlans}".format(
                        vlans=vlans
                    )
                )
        elif vlans == "none":
            payload["vlans"] = "none"
        elif vlans == "default":
            payload["vlans"] = "default"
        elif vlans.startswith("enabled") or vlans.startswith("disabled"):
            try:
                vlans_setting = vlans.split(":")[0]
                payload["vlans"] = vlans.split(":")[1].split(",")
                if vlans_setting == "disabled":
                    payload["vlans-disabled"] = True
                elif vlans_setting == "enabled":
                    payload["vlans-enabled"] = True
            except Exception:  # pylint: disable=broad-except
                return "Error: Unable to Parse vlans option: \n\tvlans={vlans}".format(
                    vlans=vlans
                )

    # determine state
    if state is not None:
        if state == "enabled":
            payload["enabled"] = True
        elif state == "disabled":
            payload["disabled"] = True

    # put to REST
    try:
        response = bigip_session.put(
            BIG_IP_URL_BASE.format(host=hostname) + f"/ltm/virtual/{name}",
            data=salt.utils.json.dumps(payload),
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def delete_virtual(hostname, username, password, name):
    """
    A function to connect to a bigip device and delete a specific virtual.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the virtual to delete

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.delete_virtual bigip admin admin my-virtual
    """

    # build session
    bigip_session = _build_session(username, password)

    # delete to REST
    try:
        response = bigip_session.delete(
            BIG_IP_URL_BASE.format(host=hostname) + f"/ltm/virtual/{name}"
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    if _load_response(response) == "":
        return True
    else:
        return _load_response(response)


def list_monitor(
    hostname,
    username,
    password,
    monitor_type,
    name=None,
):
    """
    A function to connect to a bigip device and list an existing monitor.  If no name is provided than all
    monitors of the specified type will be listed.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    monitor_type
        The type of monitor(s) to list
    name
        The name of the monitor to list

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.list_monitor bigip admin admin http my-http-monitor

    """

    # build sessions
    bigip_session = _build_session(username, password)

    # get to REST
    try:
        if name:
            response = bigip_session.get(
                BIG_IP_URL_BASE.format(host=hostname)
                + "/ltm/monitor/{type}/{name}?expandSubcollections=true".format(
                    type=monitor_type, name=name
                )
            )
        else:
            response = bigip_session.get(
                BIG_IP_URL_BASE.format(host=hostname) + f"/ltm/monitor/{monitor_type}"
            )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def create_monitor(hostname, username, password, monitor_type, name, **kwargs):
    """
    A function to connect to a bigip device and create a monitor.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    monitor_type
        The type of monitor to create
    name
        The name of the monitor to create
    kwargs
        Consult F5 BIGIP user guide for specific options for each monitor type.
        Typically, tmsh arg names are used.

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.create_monitor bigip admin admin http my-http-monitor timeout=10 interval=5
    """

    # build session
    bigip_session = _build_session(username, password)

    # construct the payload
    payload = {}
    payload["name"] = name

    # there's a ton of different monitors and a ton of options for each type of monitor.
    # this logic relies that the end user knows which options are meant for which monitor types
    for key, value in kwargs.items():
        if not key.startswith("__"):
            if key not in ["hostname", "username", "password", "type"]:
                key = key.replace("_", "-")
                payload[key] = value

    # post to REST
    try:
        response = bigip_session.post(
            BIG_IP_URL_BASE.format(host=hostname) + f"/ltm/monitor/{monitor_type}",
            data=salt.utils.json.dumps(payload),
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def modify_monitor(hostname, username, password, monitor_type, name, **kwargs):
    """
    A function to connect to a bigip device and modify an existing monitor.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    monitor_type
        The type of monitor to modify
    name
        The name of the monitor to modify
    kwargs
        Consult F5 BIGIP user guide for specific options for each monitor type.
        Typically, tmsh arg names are used.

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.modify_monitor bigip admin admin http my-http-monitor  timout=16 interval=6

    """

    # build session
    bigip_session = _build_session(username, password)

    # construct the payload
    payload = {}

    # there's a ton of different monitors and a ton of options for each type of monitor.
    # this logic relies that the end user knows which options are meant for which monitor types
    for key, value in kwargs.items():
        if not key.startswith("__"):
            if key not in ["hostname", "username", "password", "type", "name"]:
                key = key.replace("_", "-")
                payload[key] = value

    # put to REST
    try:
        response = bigip_session.put(
            BIG_IP_URL_BASE.format(host=hostname)
            + f"/ltm/monitor/{monitor_type}/{name}",
            data=salt.utils.json.dumps(payload),
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def delete_monitor(hostname, username, password, monitor_type, name):
    """
    A function to connect to a bigip device and delete an existing monitor.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    monitor_type
        The type of monitor to delete
    name
        The name of the monitor to delete

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.delete_monitor bigip admin admin http my-http-monitor

    """

    # build sessions
    bigip_session = _build_session(username, password)

    # delete to REST
    try:
        response = bigip_session.delete(
            BIG_IP_URL_BASE.format(host=hostname)
            + f"/ltm/monitor/{monitor_type}/{name}"
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    if _load_response(response) == "":
        return True
    else:
        return _load_response(response)


def list_profile(
    hostname,
    username,
    password,
    profile_type,
    name=None,
):
    """
    A function to connect to a bigip device and list an existing profile.  If no name is provided than all
    profiles of the specified type will be listed.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    profile_type
        The type of profile(s) to list
    name
        The name of the profile to list

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.list_profile bigip admin admin http my-http-profile

    """

    # build sessions
    bigip_session = _build_session(username, password)

    # get to REST
    try:
        if name:
            response = bigip_session.get(
                BIG_IP_URL_BASE.format(host=hostname)
                + "/ltm/profile/{type}/{name}?expandSubcollections=true".format(
                    type=profile_type, name=name
                )
            )
        else:
            response = bigip_session.get(
                BIG_IP_URL_BASE.format(host=hostname) + f"/ltm/profile/{profile_type}"
            )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def create_profile(hostname, username, password, profile_type, name, **kwargs):
    r"""
    A function to connect to a bigip device and create a profile.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    profile_type
        The type of profile to create
    name
        The name of the profile to create
    kwargs
        ``[ arg=val ] ... [arg=key1:val1,key2:val2] ...``

        Consult F5 BIGIP user guide for specific options for each monitor type.
        Typically, tmsh arg names are used.

    Creating Complex Args
        Profiles can get pretty complicated in terms of the amount of possible
        config options. Use the following shorthand to create complex arguments such
        as lists, dictionaries, and lists of dictionaries. An option is also
        provided to pass raw json as well.

        lists ``[i,i,i]``:
            ``param='item1,item2,item3'``

        Dictionary ``[k:v,k:v,k,v]``:
            ``param='key-1:val-1,key-2:val2,key-3:va-3'``

        List of Dictionaries ``[k:v,k:v|k:v,k:v|k:v,k:v]``:
           ``param='key-1:val-1,key-2:val-2|key-1:val-1,key-2:val-2|key-1:val-1,key-2:val-2'``

        JSON: ``'j{ ... }j'``:
           ``cert-key-chain='j{ "default": { "cert": "default.crt", "chain": "default.crt", "key": "default.key" } }j'``

        Escaping Delimiters:
            Use ``\,`` or ``\:`` or ``\|`` to escape characters which shouldn't
            be treated as delimiters i.e. ``ciphers='DEFAULT\:!SSLv3'``

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.create_profile bigip admin admin http my-http-profile defaultsFrom='/Common/http'
        salt '*' bigip.create_profile bigip admin admin http my-http-profile defaultsFrom='/Common/http' \
            enforcement=maxHeaderCount:3200,maxRequests:10

    """

    # build session
    bigip_session = _build_session(username, password)

    # construct the payload
    payload = {}
    payload["name"] = name

    # there's a ton of different profiles and a ton of options for each type of profile.
    # this logic relies that the end user knows which options are meant for which profile types
    for key, value in kwargs.items():
        if not key.startswith("__"):
            if key not in ["hostname", "username", "password", "profile_type"]:
                key = key.replace("_", "-")

                try:
                    payload[key] = _set_value(value)
                except salt.exceptions.CommandExecutionError:
                    return "Error: Unable to Parse JSON data for parameter: {key}\n{value}".format(
                        key=key, value=value
                    )

    # post to REST
    try:
        response = bigip_session.post(
            BIG_IP_URL_BASE.format(host=hostname) + f"/ltm/profile/{profile_type}",
            data=salt.utils.json.dumps(payload),
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def modify_profile(hostname, username, password, profile_type, name, **kwargs):
    r"""
    A function to connect to a bigip device and create a profile.

    A function to connect to a bigip device and create a profile.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    profile_type
        The type of profile to create
    name
        The name of the profile to create
    kwargs
        ``[ arg=val ] ... [arg=key1:val1,key2:val2] ...``

        Consult F5 BIGIP user guide for specific options for each monitor type.
        Typically, tmsh arg names are used.

    Creating Complex Args

        Profiles can get pretty complicated in terms of the amount of possible
        config options. Use the following shorthand to create complex arguments such
        as lists, dictionaries, and lists of dictionaries. An option is also
        provided to pass raw json as well.

        lists ``[i,i,i]``:
            ``param='item1,item2,item3'``

        Dictionary ``[k:v,k:v,k,v]``:
            ``param='key-1:val-1,key-2:val2,key-3:va-3'``

        List of Dictionaries ``[k:v,k:v|k:v,k:v|k:v,k:v]``:
           ``param='key-1:val-1,key-2:val-2|key-1:val-1,key-2:val-2|key-1:val-1,key-2:val-2'``

        JSON: ``'j{ ... }j'``:
           ``cert-key-chain='j{ "default": { "cert": "default.crt", "chain": "default.crt", "key": "default.key" } }j'``

        Escaping Delimiters:
            Use ``\,`` or ``\:`` or ``\|`` to escape characters which shouldn't
            be treated as delimiters i.e. ``ciphers='DEFAULT\:!SSLv3'``

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.modify_profile bigip admin admin http my-http-profile defaultsFrom='/Common/http'

        salt '*' bigip.modify_profile bigip admin admin http my-http-profile defaultsFrom='/Common/http' \
            enforcement=maxHeaderCount:3200,maxRequests:10

        salt '*' bigip.modify_profile bigip admin admin client-ssl my-client-ssl-1 retainCertificate=false \
            ciphers='DEFAULT\:!SSLv3'
            cert_key_chain='j{ "default": { "cert": "default.crt", "chain": "default.crt", "key": "default.key" } }j'
    """

    # build session
    bigip_session = _build_session(username, password)

    # construct the payload
    payload = {}
    payload["name"] = name

    # there's a ton of different profiles and a ton of options for each type of profile.
    # this logic relies that the end user knows which options are meant for which profile types
    for key, value in kwargs.items():
        if not key.startswith("__"):
            if key not in ["hostname", "username", "password", "profile_type"]:
                key = key.replace("_", "-")

                try:
                    payload[key] = _set_value(value)
                except salt.exceptions.CommandExecutionError:
                    return "Error: Unable to Parse JSON data for parameter: {key}\n{value}".format(
                        key=key, value=value
                    )

    # put to REST
    try:
        response = bigip_session.put(
            BIG_IP_URL_BASE.format(host=hostname)
            + f"/ltm/profile/{profile_type}/{name}",
            data=salt.utils.json.dumps(payload),
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    return _load_response(response)


def delete_profile(hostname, username, password, profile_type, name):
    """
    A function to connect to a bigip device and delete an existing profile.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    profile_type
        The type of profile to delete
    name
        The name of the profile to delete

    CLI Example:

    .. code-block:: bash

        salt '*' bigip.delete_profile bigip admin admin http my-http-profile

    """

    # build sessions
    bigip_session = _build_session(username, password)

    # delete to REST
    try:
        response = bigip_session.delete(
            BIG_IP_URL_BASE.format(host=hostname)
            + f"/ltm/profile/{profile_type}/{name}"
        )
    except requests.exceptions.ConnectionError as e:
        return _load_connection_error(hostname, e)

    if _load_response(response) == "":
        return True
    else:
        return _load_response(response)



import pytest

pytestmark = [
    pytest.mark.skip_on_windows,
]


def test_salt_api(api_request, install_salt, salt_master):
    """
    Test running a command against the salt api
    """
    if install_salt.distro_id in ("ubuntu", "debian"):
        pytest.skip(
            "Package test are getting reworked in https://github.com/saltstack/salt/issues/66672"
        )

    assert salt_master.is_running()

    ret = api_request.post(
        "/run",
        data={
            "client": "local",
            "tgt": "*",
            "fun": "test.arg",
            "arg": ["foo", "bar"],
        },
    )
    assert ret["args"] == ["foo", "bar"]



"""
Utility functions used for values.

.. versionadded:: 2018.3.0
"""


def xor(*variables):
    """
    XOR definition for multiple variables
    """
    sum_ = False
    for value in variables:
        sum_ = sum_ ^ bool(value)
    return sum_



from salt.cli.ssh import SaltSSH
from tests.support.mock import MagicMock, call, patch


def test_fsclient_destroy_called(minion_opts):
    """
    Test that `salt.client.ssh.SSH.fsclient.destroy()` is called.
    """
    ssh_mock = MagicMock()
    with patch(
        "salt.utils.parsers.SaltSSHOptionParser.parse_args", return_value=MagicMock()
    ), patch("salt.client.ssh.SSH", return_value=ssh_mock):
        parser = SaltSSH()
        parser.config = minion_opts
        parser.run()
    assert ssh_mock.fsclient.mock_calls == [call.destroy()]



"""
    :copyright: Copyright 2013-2017 by the SaltStack Team, see AUTHORS for more details.
    :license: Apache 2.0, see LICENSE for more details.


    tests.support.win_installer
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Fetches the binary Windows installer
"""

import hashlib
from html.parser import HTMLParser

import requests

PREFIX = "Salt-Minion-"
REPO = "https://packages.broadcom.com/artifactory/saltproject-generic/windows/"


def latest_installer_name(arch="AMD64", **kwargs):
    """
    Create an installer file name
    """

    # This is where windows packages are found
    # Each version is in its own directory, so we need to list the directories
    # and use the last one as the latest
    html_response = requests.get(REPO, timeout=60)

    versions = []

    # Create a class so we can define how to handle the starttag
    # We're looking for a "href" in the "a" tag which is the version
    class MyHTMLParser(HTMLParser):

        def handle_starttag(self, tag, attrs):
            # Only parse the 'anchor' tag.
            if tag == "a":
                # Check the list of defined attributes.
                for name, value in attrs:
                    # If href is defined, add the value to the list of versions
                    if name == "href":
                        versions.append(value.strip("/"))

    parser = MyHTMLParser()
    parser.feed(html_response.text)
    parser.close()

    latest = versions[-1]

    return f"{PREFIX}{latest}-Py3-{arch}-Setup.exe"


def download_and_verify(fp, name, repo=REPO):
    """
    Download an installer and verify its contents.
    """
    md5 = f"{name}.md5"

    def url(x):
        return f"{repo}/{x}"

    resp = requests.get(url(md5), timeout=60)
    if resp.status_code != 200:
        raise Exception("Unable to fetch installer md5")
    installer_md5 = resp.text.strip().split()[0].lower()
    resp = requests.get(url(name), stream=True, timeout=60)
    if resp.status_code != 200:
        raise Exception("Unable to fetch installer")
    md5hsh = hashlib.md5()
    for chunk in resp.iter_content(chunk_size=1024):
        md5hsh.update(chunk)
        fp.write(chunk)
    if md5hsh.hexdigest() != installer_md5:
        raise Exception(
            "Installer's hash does not match {} != {}".format(
                md5hsh.hexdigest(), installer_md5
            )
        )



import pytest

import salt.modules.alternatives as alternatives
from tests.support.helpers import TstSuiteLoggingHandler
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {alternatives: {}}


def test_display():
    with patch.dict(alternatives.__grains__, {"os_family": "RedHat"}):
        mock = MagicMock(return_value={"retcode": 0, "stdout": "salt"})
        with patch.dict(alternatives.__salt__, {"cmd.run_all": mock}):
            solution = alternatives.display("better-world")
            assert "salt" == solution
            mock.assert_called_once_with(
                ["alternatives", "--display", "better-world"],
                python_shell=False,
                ignore_retcode=True,
            )

    with patch.dict(alternatives.__grains__, {"os_family": "Suse"}):
        mock = MagicMock(return_value={"retcode": 0, "stdout": "undoubtedly-salt"})
        with patch.dict(alternatives.__salt__, {"cmd.run_all": mock}):
            solution = alternatives.display("better-world")
            assert "undoubtedly-salt" == solution
            mock.assert_called_once_with(
                ["update-alternatives", "--display", "better-world"],
                python_shell=False,
                ignore_retcode=True,
            )

    with patch.dict(alternatives.__grains__, {"os_family": "RedHat"}):
        mock = MagicMock(
            return_value={"retcode": 1, "stdout": "salt-out", "stderr": "salt-err"}
        )
        with patch.dict(alternatives.__salt__, {"cmd.run_all": mock}):
            solution = alternatives.display("better-world")
            assert "salt-err" == solution
            mock.assert_called_once_with(
                ["alternatives", "--display", "better-world"],
                python_shell=False,
                ignore_retcode=True,
            )


def test_show_current():
    mock = MagicMock(return_value="/etc/alternatives/salt")
    with patch("salt.utils.path.readlink", mock):
        ret = alternatives.show_current("better-world")
        assert "/etc/alternatives/salt" == ret
        mock.assert_called_once_with("/etc/alternatives/better-world")

        with TstSuiteLoggingHandler() as handler:
            mock.side_effect = OSError("Hell was not found!!!")
            assert not alternatives.show_current("hell")
            mock.assert_called_with("/etc/alternatives/hell")
            assert "ERROR:alternative: hell does not exist" in handler.messages


def test_check_installed():
    mock = MagicMock(return_value="/etc/alternatives/salt")
    with patch("salt.utils.path.readlink", mock):
        assert alternatives.check_installed("better-world", "/etc/alternatives/salt")
        mock.return_value = False
        assert not alternatives.check_installed("help", "/etc/alternatives/salt")


def test_install():
    with patch.dict(alternatives.__grains__, {"os_family": "RedHat"}):
        mock = MagicMock(return_value={"retcode": 0, "stdout": "salt"})
        with patch.dict(alternatives.__salt__, {"cmd.run_all": mock}):
            solution = alternatives.install(
                "better-world", "/usr/bin/better-world", "/usr/bin/salt", 100
            )
            assert "salt" == solution
            mock.assert_called_once_with(
                [
                    "alternatives",
                    "--install",
                    "/usr/bin/better-world",
                    "better-world",
                    "/usr/bin/salt",
                    "100",
                ],
                python_shell=False,
            )

    with patch.dict(alternatives.__grains__, {"os_family": "Debian"}):
        mock = MagicMock(return_value={"retcode": 0, "stdout": "salt"})
        with patch.dict(alternatives.__salt__, {"cmd.run_all": mock}):
            solution = alternatives.install(
                "better-world", "/usr/bin/better-world", "/usr/bin/salt", 100
            )
            assert "salt" == solution
            mock.assert_called_once_with(
                [
                    "update-alternatives",
                    "--install",
                    "/usr/bin/better-world",
                    "better-world",
                    "/usr/bin/salt",
                    "100",
                ],
                python_shell=False,
            )

    with patch.dict(alternatives.__grains__, {"os_family": "RedHat"}):
        mock = MagicMock(
            return_value={"retcode": 1, "stdout": "salt-out", "stderr": "salt-err"}
        )
        with patch.dict(alternatives.__salt__, {"cmd.run_all": mock}):
            ret = alternatives.install(
                "better-world", "/usr/bin/better-world", "/usr/bin/salt", 100
            )
            assert "salt-err" == ret
            mock.assert_called_once_with(
                [
                    "alternatives",
                    "--install",
                    "/usr/bin/better-world",
                    "better-world",
                    "/usr/bin/salt",
                    "100",
                ],
                python_shell=False,
            )


def test_remove():
    with patch.dict(alternatives.__grains__, {"os_family": "RedHat"}):
        mock = MagicMock(return_value={"retcode": 0, "stdout": "salt"})
        with patch.dict(alternatives.__salt__, {"cmd.run_all": mock}):
            solution = alternatives.remove(
                "better-world",
                "/usr/bin/better-world",
            )
            assert "salt" == solution
            mock.assert_called_once_with(
                ["alternatives", "--remove", "better-world", "/usr/bin/better-world"],
                python_shell=False,
            )

    with patch.dict(alternatives.__grains__, {"os_family": "Debian"}):
        mock = MagicMock(return_value={"retcode": 0, "stdout": "salt"})
        with patch.dict(alternatives.__salt__, {"cmd.run_all": mock}):
            solution = alternatives.remove(
                "better-world",
                "/usr/bin/better-world",
            )
            assert "salt" == solution
            mock.assert_called_once_with(
                [
                    "update-alternatives",
                    "--remove",
                    "better-world",
                    "/usr/bin/better-world",
                ],
                python_shell=False,
            )

    with patch.dict(alternatives.__grains__, {"os_family": "RedHat"}):
        mock = MagicMock(
            return_value={"retcode": 1, "stdout": "salt-out", "stderr": "salt-err"}
        )
        with patch.dict(alternatives.__salt__, {"cmd.run_all": mock}):
            solution = alternatives.remove(
                "better-world",
                "/usr/bin/better-world",
            )
            assert "salt-err" == solution
            mock.assert_called_once_with(
                ["alternatives", "--remove", "better-world", "/usr/bin/better-world"],
                python_shell=False,
            )


ALTERNATIVE_QUERY_EDITOR = """\
Name: editor
Link: /usr/bin/editor
Slaves:
 editor.1.gz /usr/share/man/man1/editor.1.gz
 editor.da.1.gz /usr/share/man/da/man1/editor.1.gz
 editor.de.1.gz /usr/share/man/de/man1/editor.1.gz
 editor.fr.1.gz /usr/share/man/fr/man1/editor.1.gz
 editor.it.1.gz /usr/share/man/it/man1/editor.1.gz
 editor.ja.1.gz /usr/share/man/ja/man1/editor.1.gz
 editor.pl.1.gz /usr/share/man/pl/man1/editor.1.gz
 editor.ru.1.gz /usr/share/man/ru/man1/editor.1.gz
Status: manual
Best: /bin/nano
Value: /usr/bin/vim.basic

Alternative: /bin/nano
Priority: 40
Slaves:
 editor.1.gz /usr/share/man/man1/nano.1.gz

Alternative: /usr/bin/mcedit
Priority: 25
Slaves:
 editor.1.gz /usr/share/man/man1/mcedit.1.gz

Alternative: /usr/bin/vim.basic
Priority: 30
Slaves:
 editor.1.gz /usr/share/man/man1/vim.1.gz
 editor.da.1.gz /usr/share/man/da/man1/vim.1.gz
 editor.de.1.gz /usr/share/man/de/man1/vim.1.gz
 editor.fr.1.gz /usr/share/man/fr/man1/vim.1.gz
 editor.it.1.gz /usr/share/man/it/man1/vim.1.gz
 editor.ja.1.gz /usr/share/man/ja/man1/vim.1.gz
 editor.pl.1.gz /usr/share/man/pl/man1/vim.1.gz
 editor.ru.1.gz /usr/share/man/ru/man1/vim.1.gz
"""


def test_show_link_debian():
    """Test alternatives.show_link on Debian 10."""
    run_all_mock = MagicMock(
        return_value={"retcode": 0, "stderr": "", "stdout": ALTERNATIVE_QUERY_EDITOR}
    )
    with patch.dict(alternatives.__grains__, {"os_family": "Debian"}), patch.dict(
        alternatives.__salt__, {"cmd.run_all": run_all_mock}
    ):
        assert alternatives.show_link("editor") == "/usr/bin/editor"


def test_show_link_redhat():
    """Test alternatives.show_link on CentOS 8."""
    ld_data = b"auto\n/usr/bin/ld\n\n/usr/bin/ld.bfd\n50\n/usr/bin/ld.gold\n30\n"
    with patch.dict(alternatives.__grains__, {"os_family": "RedHat"}), patch(
        "salt.utils.files.fopen",
        mock_open(read_data={"/var/lib/alternatives/ld": ld_data}),
    ):
        assert alternatives.show_link("ld") == "/usr/bin/ld"


def test_show_link_suse():
    """Test alternatives.show_link on openSUSE Leap 42.3."""
    ld_data = b"auto\n/usr/bin/ld\n\n/usr/bin/ld.bfd\n2\n\n"
    with patch.dict(alternatives.__grains__, {"os_family": "Suse"}), patch(
        "salt.utils.files.fopen",
        mock_open(read_data={"/var/lib/rpm/alternatives/ld": ld_data}),
    ):
        assert alternatives.show_link("ld") == "/usr/bin/ld"



import salt.pillar.sqlite3 as sqlite3


def test_001_extract_queries_list():
    return_data = sqlite3.SQLite3ExtPillar()
    args, kwargs = (
        [
            "SELECT blah",
            "SELECT blah2",
            ("SELECT blah3",),
            ("SELECT blah4", 2),
            {"query": "SELECT blah5"},
            {"query": "SELECT blah6", "depth": 2},
            {"query": "SELECT blah7", "as_list": True},
            {"query": "SELECT blah8", "with_lists": "1"},
            {"query": "SELECT blah9", "with_lists": "1,2"},
        ],
        {},
    )
    qbuffer = return_data.extract_queries(args, kwargs)
    assert [
        [
            None,
            {
                "query": "SELECT blah",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah2",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah3",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah4",
                "depth": 2,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah5",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah6",
                "depth": 2,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah7",
                "depth": 0,
                "as_list": True,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah8",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": [1],
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah9",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": [1, 2],
                "ignore_null": False,
            },
        ],
    ] == qbuffer


def test_002_extract_queries_kwarg():
    return_data = sqlite3.SQLite3ExtPillar()
    args, kwargs = (
        [],
        {
            "1": "SELECT blah",
            "2": "SELECT blah2",
            "3": ("SELECT blah3",),
            "4": ("SELECT blah4", 2),
            "5": {"query": "SELECT blah5"},
            "6": {"query": "SELECT blah6", "depth": 2},
            "7": {"query": "SELECT blah7", "as_list": True},
        },
    )
    qbuffer = return_data.extract_queries(args, kwargs)
    assert [
        [
            "1",
            {
                "query": "SELECT blah",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            "2",
            {
                "query": "SELECT blah2",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            "3",
            {
                "query": "SELECT blah3",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            "4",
            {
                "query": "SELECT blah4",
                "depth": 2,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            "5",
            {
                "query": "SELECT blah5",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            "6",
            {
                "query": "SELECT blah6",
                "depth": 2,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            "7",
            {
                "query": "SELECT blah7",
                "depth": 0,
                "as_list": True,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
    ] == qbuffer


def test_003_extract_queries_mixed():
    return_data = sqlite3.SQLite3ExtPillar()
    args, kwargs = (
        [
            "SELECT blah1",
            ("SELECT blah2", 2),
            {"query": "SELECT blah3", "as_list": True},
        ],
        {
            "1": "SELECT blah1",
            "2": ("SELECT blah2", 2),
            "3": {"query": "SELECT blah3", "as_list": True},
        },
    )
    qbuffer = return_data.extract_queries(args, kwargs)
    assert [
        [
            None,
            {
                "query": "SELECT blah1",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah2",
                "depth": 2,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah3",
                "depth": 0,
                "as_list": True,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            "1",
            {
                "query": "SELECT blah1",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            "2",
            {
                "query": "SELECT blah2",
                "depth": 2,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            "3",
            {
                "query": "SELECT blah3",
                "depth": 0,
                "as_list": True,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
    ] == qbuffer


def test_004_extract_queries_bogus_list():
    # This test is specifically checking that empty queries are dropped
    return_data = sqlite3.SQLite3ExtPillar()
    args, kwargs = (
        [
            "SELECT blah",
            "",
            "SELECT blah2",
            ("SELECT blah3",),
            ("",),
            ("SELECT blah4", 2),
            tuple(),
            ("SELECT blah5",),
            {"query": "SELECT blah6"},
            {"query": ""},
            {"query": "SELECT blah7", "depth": 2},
            {"not_a_query": "in sight!"},
            {"query": "SELECT blah8", "as_list": True},
        ],
        {},
    )
    qbuffer = return_data.extract_queries(args, kwargs)
    assert [
        [
            None,
            {
                "query": "SELECT blah",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah2",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah3",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah4",
                "depth": 2,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah5",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah6",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah7",
                "depth": 2,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            None,
            {
                "query": "SELECT blah8",
                "depth": 0,
                "as_list": True,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
    ] == qbuffer


def test_005_extract_queries_bogus_kwargs():
    # this test is cut down as most of the path matches test_*_bogus_list
    return_data = sqlite3.SQLite3ExtPillar()
    args, kwargs = [], {"1": "SELECT blah", "2": "", "3": "SELECT blah2"}
    qbuffer = return_data.extract_queries(args, kwargs)
    assert [
        [
            "1",
            {
                "query": "SELECT blah",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
        [
            "3",
            {
                "query": "SELECT blah2",
                "depth": 0,
                "as_list": False,
                "as_json": False,
                "with_lists": None,
                "ignore_null": False,
            },
        ],
    ] == qbuffer


def test_011_enter_root():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.enter_root("test")
    assert return_data.result["test"] == return_data.focus
    return_data.enter_root(None)
    assert return_data.result == return_data.focus


def test_021_process_fields():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b"], 0)
    assert return_data.num_fields == 2
    assert return_data.depth == 1
    return_data.process_fields(["a", "b"], 2)
    assert return_data.num_fields == 2
    assert return_data.depth == 1
    return_data.process_fields(["a", "b", "c", "d"], 0)
    assert return_data.num_fields == 4
    assert return_data.depth == 3
    return_data.process_fields(["a", "b", "c", "d"], 1)
    assert return_data.num_fields == 4
    assert return_data.depth == 1
    return_data.process_fields(["a", "b", "c", "d"], 2)
    assert return_data.num_fields == 4
    assert return_data.depth == 2
    return_data.process_fields(["a", "b", "c", "d"], 3)
    assert return_data.num_fields == 4
    assert return_data.depth == 3
    return_data.process_fields(["a", "b", "c", "d"], 4)
    assert return_data.num_fields == 4
    assert return_data.depth == 3


def test_111_process_results_legacy():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b"], 0)
    return_data.with_lists = []
    return_data.process_results([[1, 2]])
    assert {1: 2} == return_data.result


def test_112_process_results_legacy_multiple():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b"], 0)
    return_data.with_lists = []
    return_data.process_results([[1, 2], [3, 4], [5, 6]])
    assert {1: 2, 3: 4, 5: 6} == return_data.result


def test_121_process_results_depth_0():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_results([[1, 2, 3, 4], [5, 6, 7, 8]])
    assert {1: {2: {3: 4}}, 5: {6: {7: 8}}} == return_data.result


def test_122_process_results_depth_1():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b", "c", "d"], 1)
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_results([[1, 2, 3, 4], [5, 6, 7, 8]])
    assert {
        1: {"b": 2, "c": 3, "d": 4},
        5: {"b": 6, "c": 7, "d": 8},
    } == return_data.result


def test_123_process_results_depth_2():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b", "c", "d"], 2)
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_results([[1, 2, 3, 4], [5, 6, 7, 8]])
    assert {1: {2: {"c": 3, "d": 4}}, 5: {6: {"c": 7, "d": 8}}} == return_data.result


def test_124_process_results_depth_3():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b", "c", "d"], 3)
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_results([[1, 2, 3, 4], [5, 6, 7, 8]])
    assert {1: {2: {3: 4}}, 5: {6: {7: 8}}} == return_data.result


def test_125_process_results_depth_4():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b", "c", "d"], 4)
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_results([[1, 2, 3, 4], [5, 6, 7, 8]])
    assert {1: {2: {3: 4}}, 5: {6: {7: 8}}} == return_data.result


def test_131_process_results_overwrite_legacy_multiple():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b"], 0)
    return_data.with_lists = []
    return_data.process_results([[1, 2], [3, 4], [1, 6]])
    assert {1: 6, 3: 4} == return_data.result


def test_132_process_results_merge_depth_0():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_results([[1, 2, 3, 4], [1, 6, 7, 8]])
    assert {1: {2: {3: 4}, 6: {7: 8}}} == return_data.result


def test_133_process_results_overwrite_depth_0():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_results([[1, 2, 3, 4], [1, 2, 3, 8]])
    assert {1: {2: {3: 8}}} == return_data.result


def test_134_process_results_deepmerge_depth_0():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_results([[1, 2, 3, 4], [1, 2, 7, 8]])
    assert {1: {2: {3: 4, 7: 8}}} == return_data.result


def test_135_process_results_overwrite_depth_1():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b", "c", "d"], 1)
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_results([[1, 2, 3, 4], [1, 6, 7, 8]])
    assert {1: {"b": 6, "c": 7, "d": 8}} == return_data.result


def test_136_process_results_merge_depth_2():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b", "c", "d"], 2)
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_results([[1, 2, 3, 4], [1, 6, 7, 8]])
    assert {1: {2: {"c": 3, "d": 4}, 6: {"c": 7, "d": 8}}} == return_data.result


def test_137_process_results_overwrite_depth_2():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b", "c", "d"], 2)
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_results([[1, 2, 3, 4], [1, 2, 7, 8]])
    assert {1: {2: {"c": 7, "d": 8}}} == return_data.result


def test_201_process_results_complexity_multiresults():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b", "c", "d"], 2)
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_results([[1, 2, 3, 4]])
    return_data.process_results([[1, 2, 7, 8]])
    assert {1: {2: {"c": 7, "d": 8}}} == return_data.result


def test_202_process_results_complexity_as_list():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b", "c", "d"], 2)
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.as_list = True
    return_data.process_results([[1, 2, 3, 4]])
    return_data.process_results([[1, 2, 7, 8]])
    assert {1: {2: {"c": [3, 7], "d": [4, 8]}}} == return_data.result


def test_203_process_results_complexity_as_list_deeper():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.as_list = True
    return_data.process_results([[1, 2, 3, 4]])
    return_data.process_results([[1, 2, 3, 8]])
    assert {1: {2: {3: [4, 8]}}} == return_data.result


def test_204_process_results_complexity_as_list_mismatch_depth():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.as_list = True
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.process_results([[1, 2, 3, 4]])
    return_data.process_results([[1, 2, 3, 5]])
    return_data.process_fields(["a", "b", "c", "d", "e"], 0)
    return_data.process_results([[1, 2, 3, 6, 7]])
    assert {1: {2: {3: [4, 5, {6: 7}]}}} == return_data.result


def test_205_process_results_complexity_as_list_mismatch_depth_reversed():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.as_list = True
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_fields(["a", "b", "c", "d", "e"], 0)
    return_data.process_results([[1, 2, 3, 6, 7]])
    return_data.process_results([[1, 2, 3, 8, 9]])
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.process_results([[1, 2, 3, 4]])
    return_data.process_results([[1, 2, 3, 5]])
    assert {1: {2: {3: [{6: 7, 8: 9}, 4, 5]}}} == return_data.result


def test_206_process_results_complexity_as_list_mismatch_depth_weird_order():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.as_list = True
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_fields(["a", "b", "c", "d", "e"], 0)
    return_data.process_results([[1, 2, 3, 6, 7]])
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.process_results([[1, 2, 3, 4]])
    return_data.process_fields(["a", "b", "c", "d", "e"], 0)
    return_data.process_results([[1, 2, 3, 8, 9]])
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.process_results([[1, 2, 3, 5]])
    assert {1: {2: {3: [{6: 7}, 4, {8: 9}, 5]}}} == return_data.result


def test_207_process_results_complexity_collision_mismatch_depth():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.as_list = False
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.process_results([[1, 2, 3, 4]])
    return_data.process_results([[1, 2, 3, 5]])
    return_data.process_fields(["a", "b", "c", "d", "e"], 0)
    return_data.process_results([[1, 2, 3, 6, 7]])
    assert {1: {2: {3: {6: 7}}}} == return_data.result


def test_208_process_results_complexity_collision_mismatch_depth_reversed():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.as_list = False
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_fields(["a", "b", "c", "d", "e"], 0)
    return_data.process_results([[1, 2, 3, 6, 7]])
    return_data.process_results([[1, 2, 3, 8, 9]])
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.process_results([[1, 2, 3, 4]])
    return_data.process_results([[1, 2, 3, 5]])
    assert {1: {2: {3: 5}}} == return_data.result


def test_209_process_results_complexity_collision_mismatch_depth_weird_order():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.as_list = False
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_fields(["a", "b", "c", "d", "e"], 0)
    return_data.process_results([[1, 2, 3, 6, 7]])
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.process_results([[1, 2, 3, 4]])
    return_data.process_fields(["a", "b", "c", "d", "e"], 0)
    return_data.process_results([[1, 2, 3, 8, 9]])
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.process_results([[1, 2, 3, 5]])
    assert {1: {2: {3: 5}}} == return_data.result


def test_20A_process_results_complexity_as_list_vary():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.as_list = True
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_fields(["a", "b", "c", "d", "e"], 0)
    return_data.process_results([[1, 2, 3, 6, 7]])
    return_data.process_results([[1, 2, 3, 8, 9]])
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.process_results([[1, 2, 3, 4]])
    return_data.as_list = False
    return_data.process_results([[1, 2, 3, 5]])
    assert {1: {2: {3: 5}}} == return_data.result


def test_207_process_results_complexity_roots_collision():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.as_list = False
    return_data.with_lists = []
    return_data.enter_root(None)
    return_data.process_fields(["a", "b", "c", "d"], 0)
    return_data.process_results([[1, 2, 3, 4]])
    return_data.enter_root(1)
    return_data.process_results([[5, 6, 7, 8]])
    assert {1: {5: {6: {7: 8}}}} == return_data.result


def test_301_process_results_with_lists():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.as_list = False
    return_data.with_lists = [1, 3]
    return_data.enter_root(None)
    return_data.process_fields(["a", "b", "c", "d", "e", "v"], 0)
    return_data.process_results(
        [
            ["a", "b", "c", "d", "e", 1],
            ["a", "b", "c", "f", "g", 2],
            ["a", "z", "h", "y", "j", 3],
            ["a", "z", "h", "y", "k", 4],
        ]
    )
    assert sorted(
        {"a": [{"c": [{"e": 1}, {"g": 2}]}, {"h": [{"j": 3, "k": 4}]}]}
    ) == sorted(return_data.result)


def test_302_process_results_with_lists_consecutive():
    return_data = sqlite3.SQLite3ExtPillar()
    return_data.as_list = False
    return_data.with_lists = [1, 2, 3]
    return_data.enter_root(None)
    return_data.process_fields(["a", "b", "c", "d", "e", "v"], 0)
    return_data.process_results(
        [
            ["a", "b", "c", "d", "e", 1],
            ["a", "b", "c", "f", "g", 2],
            ["a", "z", "h", "y", "j", 3],
            ["a", "z", "h", "y", "k", 4],
        ]
    )
    assert sorted({"a": [[[{"e": 1}, {"g": 2}]], [[{"j": 3, "k": 4}]]]}) == sorted(
        return_data.result
    )



"""
Beacon that fires events on image import/delete.

.. code-block:: yaml

    ## minimal
    # - check for new images every 1 second (salt default)
    # - does not send events at startup
    beacons:
      imgadm: []

    ## standard
    # - check for new images every 60 seconds
    # - send import events at startup for all images
    beacons:
      imgadm:
        - interval: 60
        - startup_import_event: True
"""

import logging

import salt.utils.beacons

__virtualname__ = "imgadm"

IMGADM_STATE = {
    "first_run": True,
    "images": [],
}

log = logging.getLogger(__name__)


def __virtual__():
    """
    Provides imgadm beacon on SmartOS
    """
    if "imgadm.list" in __salt__:
        return True
    else:
        err_msg = "Only available on SmartOS compute nodes."
        log.error("Unable to load %s beacon: %s", __virtualname__, err_msg)
        return False, err_msg


def validate(config):
    """
    Validate the beacon configuration
    """
    vcfg_ret = True
    vcfg_msg = "Valid beacon configuration"

    if not isinstance(config, list):
        vcfg_ret = False
        vcfg_msg = "Configuration for imgadm beacon must be a list!"

    return vcfg_ret, vcfg_msg


def beacon(config):
    """
    Poll imgadm and compare available images
    """
    ret = []

    # NOTE: lookup current images
    current_images = __salt__["imgadm.list"](verbose=True)

    # NOTE: apply configuration
    if IMGADM_STATE["first_run"]:
        log.info("Applying configuration for imgadm beacon")

        config = salt.utils.beacons.list_to_dict(config)

        if "startup_import_event" not in config or not config["startup_import_event"]:
            IMGADM_STATE["images"] = current_images

    # NOTE: import events
    for uuid in current_images:
        event = {}
        if uuid not in IMGADM_STATE["images"]:
            event["tag"] = f"imported/{uuid}"
            for label in current_images[uuid]:
                event[label] = current_images[uuid][label]

        if event:
            ret.append(event)

    # NOTE: delete events
    for uuid in IMGADM_STATE["images"]:
        event = {}
        if uuid not in current_images:
            event["tag"] = f"deleted/{uuid}"
            for label in IMGADM_STATE["images"][uuid]:
                event[label] = IMGADM_STATE["images"][uuid][label]

        if event:
            ret.append(event)

    # NOTE: update stored state
    IMGADM_STATE["images"] = current_images

    # NOTE: disable first_run
    if IMGADM_STATE["first_run"]:
        IMGADM_STATE["first_run"] = False

    return ret



import logging
import subprocess

import pytest
from _pytest.outcomes import Failed
from _pytest.pytester import LineMatcher

log = logging.getLogger(__name__)


def test_logging_and_state_output_order(salt_master, salt_minion, salt_cli, tmp_path):
    """
    This tests for any regressions for this issue:
    https://github.com/saltstack/salt/issues/62005
    """
    target_path = tmp_path / "file-target.txt"
    sls_name = "file-target"
    sls_contents = """
    add_contents_pillar_sls:
      file.managed:
        - name: {}
        - contents: foo
    """.format(
        target_path
    )
    sls_tempfile = salt_master.state_tree.base.temp_file(
        f"{sls_name}.sls", sls_contents
    )
    with sls_tempfile:
        # Get the command line to use
        cmdline = salt_cli.cmdline(
            "-ldebug", "state.sls", sls_name, minion_tgt=salt_minion.id
        )
        assert cmdline
        # Use subprocess.run since we want the output of stdout(state output) and stderr(logging)
        # mixed so we can check for the order
        ret = subprocess.run(
            cmdline,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=False,
            shell=False,
            text=True,
        )
        assert ret.stdout
        assert not ret.stderr
        log.debug("STDOUT:\n>>>>>>>>\n%s\n<<<<<<<\n", ret.stdout)
        matcher = LineMatcher(ret.stdout.splitlines())
        assert ret.returncode == 0
        assert target_path.is_file()
        # Check for proper order of state output and logging
        try:
            # This output order should not match and should trigger a _pytest.outcomes.Failed exception
            matcher.fnmatch_lines(
                [
                    f'"{salt_minion.id}":*',
                    '"file_*',
                    "*Reading configuration from*",
                ]
            )
        except Failed:
            # We caught the expected failure regarding the output matching above,
            # nonetheless, let's confirm proper output order
            matcher.fnmatch_lines(
                [
                    # Confirm we have logging going on...
                    "*Reading configuration from*",
                    # And that after logging, we have the state output
                    f'"{salt_minion.id}":*',
                    '"file_*',
                ]
            )
        else:
            pytest.fail("The state and logging output order is wrong")



import pytest
from requests.models import Response

import salt.modules.swarm as swarm
from tests.support.mock import DEFAULT, MagicMock, patch

HAS_DOCKER = False
try:
    import docker

    HAS_DOCKER = True
except ImportError:
    HAS_DOCKER = False

pytestmark = [
    pytest.mark.skipif(
        HAS_DOCKER is False, reason="The docker python sdk may not be installed"
    ),
]


@pytest.fixture
def configure_loader_modules():
    return {swarm: {"__context__": {}}}


@pytest.fixture
def fake_context_client(setup_loader_mock):
    fake_swarm_client = MagicMock()
    patch_context = patch.dict(
        swarm.__context__, {"client": fake_swarm_client, "server_name": "test swarm"}
    )
    patch_swarm_token = patch(
        "salt.modules.swarm.swarm_tokens", autospec=True, return_value="mocked_token"
    )
    with patch_context, patch_swarm_token:
        yield fake_swarm_client


def test_when_swarm_init_is_called_with_the_same_information_twice_it_should_return_the_docker_error(
    fake_context_client,
):
    error_response = Response()
    error_response._content = (
        b'{"message":"This node is already part of a swarm. Use \\"docker swarm'
        b' leave\\" to leave this swarm and join another one."}\n'
    )
    error_response.status_code = 503
    error_response.reason = "Service Unavailable"
    swarm_error_message = (
        'This node is already part of a swarm. Use "docker swarm leave" to leave this'
        " swarm and join another one."
    )
    fake_context_client.swarm.init.side_effect = [
        DEFAULT,
        docker.errors.APIError(
            swarm_error_message,
            response=error_response,
            explanation=swarm_error_message,
        ),
    ]

    expected_good_result = {
        "Comment": (
            "Docker swarm has been initialized on test swarm and the worker/manager"
            " Join token is below"
        ),
        "Tokens": "mocked_token",
    }
    expected_bad_result = {"Comment": swarm_error_message, "result": False}

    first_result = swarm.swarm_init("127.0.0.1", "0.0.0.0", False)
    second_result = swarm.swarm_init("127.0.0.1", "0.0.0.0", False)

    assert first_result == expected_good_result
    assert second_result == expected_bad_result



"""
    :codeauthor: Wayne Werner <wwerner@saltstack.com>
"""

# Import the future

import os
import tempfile

# Salt Libs
import salt.modules.cmdmod as cmd
import salt.modules.file as file
import salt.modules.tls as tls
import salt.utils.files as files
import salt.utils.stringutils as stringutils

# Testing libs
from tests.support.case import ModuleCase
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock
from tests.support.runtests import RUNTIME_VARS


class TLSModuleTest(ModuleCase, LoaderModuleMockMixin):
    """
    Tests for salt.modules.tls
    """

    def setup_loader_modules(self):
        opts = {
            "cachedir": os.path.join(RUNTIME_VARS.TMP, "cache"),
            "test": True,
        }
        return {
            tls: {
                "__salt__": {
                    "config.option": MagicMock(return_value=self.tempdir),
                    "cmd.retcode": cmd.retcode,
                    "pillar.get": MagicMock(return_value=False),
                    "file.replace": file.replace,
                },
                "__opts__": opts,
            },
            file: {
                "__utils__": {
                    "files.is_text": files.is_text,
                    "stringutils.get_diff": stringutils.get_diff,
                },
                "__opts__": opts,
            },
        }

    @classmethod
    def setUpClass(cls):
        cls.ca_name = "roscivs"
        cls.tempdir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

    def test_ca_exists_should_be_False_before_ca_is_created(self):
        self.assertFalse(tls.ca_exists(self.ca_name))

    def test_ca_exists_should_be_True_after_ca_is_created(self):
        tls.create_ca(self.ca_name)
        self.assertTrue(tls.ca_exists(self.ca_name))

    def test_creating_csr_should_fail_with_no_ca(self):
        expected_message = (
            'Certificate for CA named "bad_ca" does not exist, please create it first.'
        )
        self.assertEqual(tls.create_csr(ca_name="bad_ca"), expected_message)

    def test_with_existing_ca_signing_csr_should_produce_valid_cert(self):
        print("Revoked should not be here")
        empty_crl_filename = os.path.join(self.tempdir, "empty.crl")
        tls.create_ca(self.ca_name)
        tls.create_csr(
            ca_name=self.ca_name,
            CN="testing.localhost",
        )
        tls.create_ca_signed_cert(
            ca_name=self.ca_name,
            CN="testing.localhost",
        )
        tls.create_empty_crl(
            ca_name=self.ca_name,
            crl_file=empty_crl_filename,
        )
        ret = tls.validate(
            cert=os.path.join(
                self.tempdir,
                self.ca_name,
                "certs",
                "testing.localhost.crt",
            ),
            ca_name=self.ca_name,
            crl_file=empty_crl_filename,
        )
        print("not there")
        self.assertTrue(ret["valid"], ret.get("error"))

    def test_revoked_cert_should_return_False_from_validate(self):
        revoked_crl_filename = os.path.join(self.tempdir, "revoked.crl")
        tls.create_ca(self.ca_name)
        tls.create_csr(
            ca_name=self.ca_name,
            CN="testing.bad.localhost",
        )
        tls.create_ca_signed_cert(
            ca_name=self.ca_name,
            CN="testing.bad.localhost",
        )
        tls.create_empty_crl(
            ca_name=self.ca_name,
            crl_file=revoked_crl_filename,
        )
        tls.revoke_cert(
            ca_name=self.ca_name,
            CN="testing.bad.localhost",
            crl_file=revoked_crl_filename,
        )
        self.assertFalse(
            tls.validate(
                cert=os.path.join(
                    self.tempdir,
                    self.ca_name,
                    "certs",
                    "testing.bad.localhost.crt",
                ),
                ca_name=self.ca_name,
                crl_file=revoked_crl_filename,
            )["valid"]
        )

    def test_validating_revoked_cert_with_no_crl_file_should_return_False(self):
        revoked_crl_filename = None
        tls.create_ca(self.ca_name)
        tls.create_csr(
            ca_name=self.ca_name,
            CN="testing.bad.localhost",
        )
        tls.create_ca_signed_cert(
            ca_name=self.ca_name,
            CN="testing.bad.localhost",
        )
        tls.create_empty_crl(
            ca_name=self.ca_name,
            crl_file=revoked_crl_filename,
        )
        tls.revoke_cert(
            ca_name=self.ca_name,
            CN="testing.bad.localhost",
            crl_file=revoked_crl_filename,
        )
        self.assertFalse(
            tls.validate(
                cert=os.path.join(
                    self.tempdir,
                    self.ca_name,
                    "certs",
                    "testing.bad.localhost.crt",
                ),
                ca_name=self.ca_name,
                crl_file=revoked_crl_filename,
            )["valid"]
        )



import os

import pytest

import salt.loader
import salt.modules.config
import salt.modules.cp
import salt.modules.slsutil
import salt.utils.files
from tests.support.mock import MagicMock

try:
    import salt.modules.yaml
    import salt.utils.yamllint

    YAMLLINT_AVAILABLE = salt.utils.yamllint.has_yamllint()
except ImportError:
    YAMLLINT_AVAILABLE = False


pytestmark = [
    pytest.mark.skip_on_windows(reason="Dam line breaks"),
    pytest.mark.skipif(
        YAMLLINT_AVAILABLE is False, reason="The 'yammllint' pacakge is not available"
    ),
]


@pytest.fixture
def configure_loader_modules(minion_opts):
    with pytest.helpers.temp_file(
        "testyaml.yaml", contents="key:\n  value\n"
    ) as cached_file:
        yield {
            salt.modules.yaml: {
                "__salt__": {
                    "config.get": salt.modules.config.get,
                    "cp.cache_file": MagicMock(
                        salt.modules.cp.cache_file,
                        autospec=True,
                        return_value=str(cached_file),
                    ),
                    "slsutil.renderer": MagicMock(
                        salt.modules.slsutil.renderer,
                        autospec=True,
                        return_value=f"key: value{os.linesep}",
                    ),
                },
                "__opts__": minion_opts,
                "__utils__": {
                    "files.fopen": salt.utils.files.fopen,
                    "yamllint.lint": salt.utils.yamllint.lint,
                },
            },
            salt.modules.config: {
                "__opts__": minion_opts,
            },
        }


def test_lint_yaml():
    """
    ensure that we can lint from the yaml lint utils
    """
    assert salt.modules.yaml.lint("salt://test/test.sls") == {
        "problems": [],
        "source": f"key:{os.linesep}  value{os.linesep}",
    }


def test_lint_pre_render():
    assert salt.modules.yaml.lint("salt://test.test.sls", pre_render="jinja") == {
        "problems": [],
        "source": f"key: value{os.linesep}",
    }


def test_yamllint_virtual():
    assert salt.modules.yaml.__virtual__() == "yaml"



"""
mac_power tests
"""

import pytest

import salt.modules.mac_power as mac_power
from salt.exceptions import SaltInvocationError


@pytest.fixture
def configure_loader_modules():
    return {mac_power: {}}


def test_validate_sleep_valid_number():
    """
    test _validate_sleep function with valid number
    """
    assert mac_power._validate_sleep(179) == 179


def test_validate_sleep_invalid_number():
    """
    test _validate_sleep function with invalid number
    """
    pytest.raises(SaltInvocationError, mac_power._validate_sleep, 181)


def test_validate_sleep_valid_string():
    """
    test _validate_sleep function with valid string
    """
    assert mac_power._validate_sleep("never") == "Never"
    assert mac_power._validate_sleep("off") == "Never"


def test_validate_sleep_invalid_string():
    """
    test _validate_sleep function with invalid string
    """
    pytest.raises(SaltInvocationError, mac_power._validate_sleep, "bob")


def test_validate_sleep_bool_true():
    """
    test _validate_sleep function with True
    """
    pytest.raises(SaltInvocationError, mac_power._validate_sleep, True)


def test_validate_sleep_bool_false():
    """
    test _validate_sleep function with False
    """
    assert mac_power._validate_sleep(False) == "Never"


def test_validate_sleep_unexpected():
    """
    test _validate_sleep function with True
    """
    pytest.raises(SaltInvocationError, mac_power._validate_sleep, 172.7)
